"""Template jinja tags."""

import collections as py_collections
import itertools
from datetime import datetime
import jinja2
from babel.messages import catalog as babel_catalog
from grow.collections import collection as collection_lib
from grow.common import structures
from grow.common import utils
from grow.pods import errors
from grow.translations import locales as locales_lib


class Menu(object):
    """Helper class for creating navigation menus."""

    def __init__(self):
        self._items = py_collections.OrderedDict()

    def build(self, nodes):
        """Builds the menu from the set of nodes."""
        self._recursive_build(self._items, None, nodes)

    def items(self):
        """Iterate through items."""
        return self._items.items()

    def _recursive_build(self, tree, parent, nodes):
        children = [n for n in nodes if n.parent == parent]
        for child in children:
            tree[child] = py_collections.OrderedDict()
            self._recursive_build(tree[child], child, nodes)


@jinja2.contextfunction
def _gettext_alias(__context, *args, **kwargs):
    return __context.call(__context.resolve('gettext'), *args, **kwargs)


def _read_csv(pod, pod_path, locale):
    """Reads a csv file using a cache."""
    file_cache = pod.podcache.file_cache
    contents = file_cache.get(pod_path, locale=locale)
    if contents is None:
        contents = pod.read_csv(pod_path, locale=locale)
        contents = pod.untag(contents, locale=locale)
        file_cache.add(pod_path, contents, locale=locale)
    return contents


def _read_json(pod, pod_path, locale):
    """Reads a json file using a cache."""
    file_cache = pod.podcache.file_cache
    contents = file_cache.get(pod_path)
    if contents is None:
        contents = pod.read_json(pod_path)
        contents = pod.untag(contents, locale=locale)
        file_cache.add(pod_path, contents, locale=locale)
    return contents


def _read_yaml(pod, pod_path, locale):
    """Reads a yaml file using a cache."""
    file_cache = pod.podcache.file_cache
    contents = file_cache.get(pod_path, locale)
    if contents is None:
        contents = pod.read_yaml(pod_path, locale=locale)
        file_cache.add(pod_path, contents, locale=locale)
    return contents


# pylint: disable=redefined-outer-name
def categories(collection=None, reverse=None, recursive=True,
               locale=utils.SENTINEL, _pod=None):
    """Retrieve catagories from the pod."""
    if isinstance(collection, collection_lib.Collection):
        collection = collection
    elif isinstance(collection, str):
        collection = _pod.get_collection(collection)
    else:
        text = '{} must be a Collection instance or a collection path, found: {}.'
        raise ValueError(text.format(collection, type(collection)))
    # Collection's categories are only used for sort order.
    docs = collection.docs(reverse=reverse, locale=locale, order_by='category',
                           recursive=recursive)
    result = []
    for category, unsorted_docs in itertools.groupby(docs, key=lambda doc: doc.category):
        sorted_docs = sorted(unsorted_docs, key=lambda doc: doc.order)
        result.append((category, sorted_docs))
    return result


@utils.memoize_tag
def collection(collection, _pod=None):
    """Retrieves a collection from the pod."""
    return _pod.get_collection(collection)


@utils.memoize_tag
def collections(collection_paths=None, _pod=None):
    """Retrieves collections from the pod."""
    return _pod.list_collections(collection_paths)


@utils.memoize_tag
def csv(path, locale=utils.SENTINEL, _pod=None):
    """Retrieves a csv file from the pod."""
    return _read_csv(_pod, path, locale=locale)


def date(datetime_obj=None, _pod=None, **kwargs):
    """Creates a date optionally from another date."""
    _from = kwargs.get('from', None)
    if datetime_obj is None:
        datetime_obj = datetime.now()
    elif isinstance(datetime_obj, str) and _from is not None:
        datetime_obj = datetime.strptime(datetime_obj, _from)
    return datetime_obj


@utils.memoize_tag
def docs(collection, _pod=None, **kwargs):
    """Retrieves docs from the pod."""
    collection = _pod.get_collection(collection)
    return list(collection.list_docs(**kwargs))


@utils.memoize_tag
def get_doc(pod_path, locale=None, _pod=None):
    """Retrieves a doc from the pod."""
    return _pod.get_doc(pod_path, locale=locale)


@utils.memoize_tag
def json(path, _pod, locale=None):
    """Retrieves a json file from the pod."""
    return _read_json(_pod, path, locale=locale)


@utils.memoize_tag
def locale_tag(code, _pod=None):
    """Parses locale from a given locale code."""
    return locales_lib.Locale.parse(code)


@utils.memoize_tag
def locales_tag(codes, _pod=None):
    """Parses locales from the given locale codes."""
    return locales_lib.Locale.parse_codes(codes)


def make_doc_gettext(doc):
    """Create a gettext function that tracks translation stats."""
    if not doc:
        return _gettext_alias

    translation_stats = doc.pod.translation_stats
    catalog = doc.pod.catalogs.get(doc.locale)

    @jinja2.contextfunction
    def gettext(__context, __string, *args, **kwargs):
        # Elegantly handle non-strings passed to gettext.
        if not isinstance(__string, str):
            return __string
        # Use the message from the catalog. If there is none (if `extract`
        # hasn't been run yet to create empty messages), create a stub message
        # used for tracking untranslated strings.
        message = catalog[__string] \
            or babel_catalog.Message(__string, locations=[(doc.pod_path, 0)])
        translation_stats.tick(
            message, doc.locale, doc.default_locale, location=doc.pod_path)
        return __context.call(__context.resolve('gettext'), __string, *args, **kwargs)
    return gettext


def make_gettext(pod, func):
    """Create a gettext function that tracks translation stats."""
    @jinja2.contextfunction
    def gettext(__context, __string, **variables):
        """Gettext and do replacement."""
        value = __context.call(func, __string)
        value = utils.safe_format(value, **variables)
        if __context.eval_ctx.autoescape:
            value = jinja2.utils.Markup(value)
        return value
    return gettext


def make_ngettext(pod, func):
    """Create a gettext function that tracks translation stats."""
    @jinja2.contextfunction
    def ngettext(__context, __singular, __plural, __num, **variables):
        """Gettext and do replacement."""
        variables.setdefault('num', __num)
        value = __context.call(func, __singular, __plural, __num)
        value = utils.safe_format(value, **variables)
        if __context.eval_ctx.autoescape:
            value = jinja2.utils.Markup(value)
        return value
    return ngettext


@utils.memoize_tag
def nav(collection=None, locale=None, _pod=None):
    """Builds a navigation object for templates."""
    collection_obj = _pod.get_collection('/content/' + collection)
    results = collection_obj.docs(order_by='order', locale=locale)
    menu = Menu()
    menu.build(results)
    return menu


@utils.memoize_tag
def static_file(path, locale=None, _pod=None):
    """Retrieves a static file from the pod."""
    return _pod.get_static(path, locale=locale)


@utils.memoize_tag
def statics(pod_path, locale=None, include_hidden=False, _pod=None):
    """Retrieves a list of statics from the pod."""
    return list(_pod.list_statics(pod_path, locale=locale, include_hidden=include_hidden))


@utils.memoize_tag
def url(pod_path, locale=None, _pod=None):
    """Retrieves a url for a given document in the pod."""
    return _pod.get_url(pod_path, locale=locale)


def wrap_locale_context(func):
    """Wraps the func with the current locale."""

    @jinja2.contextfilter
    def _locale_filter(ctx, value, *args, **kwargs):
        doc = ctx['doc']
        if not kwargs.get('locale', None):
            kwargs['locale'] = str(doc.locale)
        return func(value, *args, **kwargs)
    return _locale_filter


@utils.memoize_tag
def yaml(path, _pod, _doc=None):
    """Retrieves a yaml file from the pod."""
    locale = str(_doc.locale_safe) if _doc else None
    if '?' in path:
        path, reference = path.split('?')
        data = structures.DeepReferenceDict(_read_yaml(_pod, path, locale=locale))
        return data[reference]
    return _read_yaml(_pod, path, locale=locale)


def create_builtin_tags(pod, doc, track_dependency=None):
    """Creates standard set of tags for rendering based on the doc."""

    def _wrap(func):
        # pylint: disable=unnecessary-lambda
        return lambda *args, **kwargs: func(*args, _pod=pod, **kwargs)

    def _wrap_dependency(func):
        def _wrapper(*args, **kwargs):
            if doc and not kwargs.get('locale', None):
                # Only override the locale kwarg if it is absent. Avoid
                # replacing an explicit `locale=None` with the document's
                # current locale. The presence of a `locale=None` indicates
                # caller specifies "no locale" rather than "the current
                # document's default locale".
                if 'locale' not in kwargs:
                    kwargs['locale'] = str(doc.locale)
            included_docs = func(*args, _pod=pod, **kwargs)
            if track_dependency:
                try:
                    for included_doc in included_docs:
                        track_dependency(included_doc.pod_path)
                except TypeError:
                    # Not an interable, try it as a doc.
                    track_dependency(included_docs.pod_path)
            return included_docs
        return _wrapper

    def _wrap_dependency_path(func):
        def _wrapper(*args, **kwargs):
            if track_dependency:
                track_dependency(args[0])
            return func(*args, _pod=pod, **kwargs)
        return _wrapper

    def _wrap_doc(func):
        def _wrapper(*args, **kwargs):
            if doc:
                kwargs['_doc'] = doc
            return func(*args, **kwargs)
        return _wrapper

    return {
        'categories': _wrap(categories),
        'collection': _wrap(collection),
        'collections': _wrap(collections),
        'csv': _wrap_dependency_path(csv),
        'date': _wrap(date),
        'doc': _wrap_dependency(get_doc),
        'docs': _wrap_dependency(docs),
        'json': _wrap_dependency_path(json),
        'locale': _wrap(locale_tag),
        'locales': _wrap(locales_tag),
        'nav': _wrap(nav),
        'static': _wrap_dependency(static_file),
        'statics': _wrap_dependency(statics),
        'url': _wrap_dependency_path(url),
        'yaml': _wrap_dependency_path(_wrap_doc(yaml)),
    }


def create_builtin_globals(env, pod, locale=None):
    """Create built in global tags."""
    # Mark that we are using the newstyle gettext to avoid issues with escaping.
    env.newstyle_gettext = True
    get_gettext_func = pod.catalogs.get_gettext_translations
    gettext = make_gettext(
        pod, lambda x: get_gettext_func(locale).ugettext(x))
    ngettext = make_ngettext(
        pod, lambda s, p, n: get_gettext_func(locale).ungettext(s, p, n))
    return {
        'gettext': gettext,
        'ngettext': ngettext,
    }



"""
Parsing and manipulation of front matter elements of a document.
"""

import collections
import re
import yaml
from grow.common import untag
from grow.common import utils

BOUNDARY_REGEX = re.compile(r'^-{3,}\s*$', re.MULTILINE)
CONVERT_MESSAGE = """Document contains too many parts: {}
    Remove superfluous separators (`---`).
    Or, please run `grow convert --type content_locale_split` to help update files."""


def _update_deep(orig_dict, new_dict):
    for k in new_dict:
        if (k in orig_dict and isinstance(orig_dict[k], dict)
                and isinstance(new_dict[k], collections.Mapping)):
            _update_deep(orig_dict[k], new_dict[k])
        else:
            orig_dict[k] = new_dict[k]


class Error(Exception):
    """General document front matter error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class BadFormatError(Error, ValueError):
    """Error for bad document front matter formatting."""
    pass


class DocumentFrontMatter(object):
    """Document front matter."""

    def __init__(self, doc, raw_front_matter=None):
        self._doc = doc
        self.data = {}
        self.update_raw_front_matter(raw_front_matter)

    @staticmethod
    def split_front_matter(content, pod_path=''):
        """Takes raw document content and separates the front matter and content."""
        parts = BOUNDARY_REGEX.split(content)
        if parts[0].strip() == '':
            parts.pop(0)
        if len(parts) > 2:
            message = CONVERT_MESSAGE.format(pod_path)
            raise BadFormatError(message)
        if len(parts) == 2:
            return parts[0].strip() or None, parts[1].strip()
        return None, content.strip()

    def _load_front_matter(self, raw_front_matter):
        """
        Documents with localization need to base the front-matter off the base
        document to extend or overwrite the base fields.
        """
        # Load base locales from the raw yaml to prevent shared variables.
        for locale_path in reversed(self._doc.locale_paths[1:]):
            if self._doc.pod.file_exists(locale_path):
                locale_doc = self._doc.pod.get_doc(locale_path)
                raw_locale_front_matter = locale_doc.format.front_matter.export()
                if raw_locale_front_matter:
                    _update_deep(self.data, self._load_yaml(
                        raw_locale_front_matter))

        if raw_front_matter:
            # There should be no boundary separators in the front-matter.
            if BOUNDARY_REGEX.search(raw_front_matter):
                message = CONVERT_MESSAGE.format(
                    self._doc.pod_path)
                raise BadFormatError(message)
            self._raw_front_matter = raw_front_matter
        elif self._doc.exists:
            self._raw_front_matter, _ = DocumentFrontMatter.split_front_matter(
                self._doc.raw_content, pod_path=self._doc.pod_path)

        if self._raw_front_matter:
            new_data = self._load_yaml(self._raw_front_matter)
            if not isinstance(new_data, dict):
                raise BadFormatError(
                    'Front matter needs to be a dictionary: {} <{}>'.format(
                        self._doc.pod_path, type(new_data).__name__))
            _update_deep(self.data, new_data)

    def _load_yaml(self, raw_yaml):
        try:
            return utils.load_yaml(
                raw_yaml, doc=self._doc, pod=self._doc.pod,
                untag_params={
                    'env': untag.UntagParamRegex(self._doc.pod.env.name),
                    'locale': untag.UntagParamLocaleRegex.from_pod(
                        self._doc.pod, self._doc.collection),
                })
        except (yaml.composer.ComposerError,
                yaml.parser.ParserError,
                yaml.reader.ReaderError,
                yaml.scanner.ScannerError) as error:
            message = 'Error parsing {}: {}'.format(self._doc.pod_path, error)
            raise BadFormatError(message)

    @property
    def raw_data(self):
        if not self._raw_front_matter:
            return {}
        return utils.load_yaml(
            self._raw_front_matter, doc=self._doc, pod=self._doc.pod)

    def export(self):
        """
        Exports the front matter in a raw format.

        This allows it to be cached in a file and parsed when recontructed.
        If exported in the parsed form then any yaml constructors will have
        already run and it will fail since it cannot write the referenced
        objects in the cache file. Also the size can be much larger after the
        yaml has been parsed.
        """
        return self._raw_front_matter

    def update_fields(self, fields):
        """Update the data with new field values."""
        _update_deep(self.data, fields)
        self.update_raw_front_matter(utils.dump_yaml(self.data))

    def update_raw_front_matter(self, raw_front_matter):
        """Replace the value of the front matter using a new raw string."""
        self.data = {}
        self._raw_front_matter = None
        self._load_front_matter(raw_front_matter)



# coding: utf8
"""Base installer class."""

from grow.common import colors
from grow.sdk.installers import base_installer


EXTRA_FORMAT = '{}\n   {}'
MESSAGE_FORMAT = '[{}] {}'


class Error(Exception):
    """Base error for installers."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class Installer(object):
    """Grow installer for dependencies."""

    def __init__(self, installers, pod):
        self.installers = installers
        self.pod = pod

    @staticmethod
    def format_message(message, extras=None):
        """Format a message with support for extra lines."""
        if extras:
            for extra in extras:
                message = EXTRA_FORMAT.format(message, extra)
        return message

    @classmethod
    def failure(cls, message, extras=None):
        """Generate a formatted failure message."""
        message = cls.format_message(message, extras)
        return colors.stylize(MESSAGE_FORMAT.format('✘', message), colors.ERROR)

    @classmethod
    def pre_install(cls, message, extras=None):
        """Generate a formatted pre install message."""
        message = cls.format_message(message, extras)
        return colors.stylize(MESSAGE_FORMAT.format(' ', message), colors.SUCCESS)

    @classmethod
    def success(cls, message, extras=None):
        """Generate a formatted success message."""
        message = cls.format_message(message, extras)
        return colors.stylize(MESSAGE_FORMAT.format('✓', message), colors.SUCCESS)

    def run_installers(self):
        """Run each installer to install prerequisites."""
        for installer in self.installers:
            if not installer.should_run:
                continue

            with self.pod.profile.timer('Installer.run_installers.{}'.format(installer.KIND)):
                try:
                    installer.check_prerequisites()
                except base_installer.MissingPrerequisiteError as err:
                    self.pod.logger.error(
                        self.failure(err.message, extras=err.install_commands))
                    return False

                messages = installer.pre_install_messages
                if messages:
                    self.pod.logger.info(
                        self.pre_install(messages[0], messages[1:] if len(messages) > 1 else []))

                installer.install()

                messages = installer.post_install_messages
                if messages:
                    self.pod.logger.info(
                        self.success(messages[0], messages[1:] if len(messages) > 1 else []))
        # All installers were successful.
        return True



"""Collections contain content documents and blueprints."""

import copy
import logging
import operator
import os
import re
import sys
from grow.collections import collection_routes
from grow.common import features
from grow.common import structures
from grow.common import untag
from grow.common import utils
from grow.documents import document
from grow.documents import document_front_matter
from grow.translations import locales
from grow.pods import messages


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class BadCollectionNameError(Error, ValueError):
    pass


class BadFieldsError(Error, ValueError):
    pass


class CollectionDoesNotExistError(Error, ValueError):
    pass


class CollectionExistsError(Error):
    pass


class CollectionNotEmptyError(Error):
    pass


class NoLocalesError(Error):
    pass


class Collection(object):
    CONTENT_PATH = '/content'
    BLUEPRINT_PATH = '_blueprint.yaml'
    FEATURE_SEPARATE_ROUTING = 'separate_routing'
    IGNORE_INITIAL = ('_',)

    _content_path_regex = re.compile('^' + CONTENT_PATH + '/?')

    def __eq__(self, other):
        return (isinstance(other, Collection)
                and self.collection_path == other.collection_path)

    def __getattr__(self, name):
        # Stay back from the recursion loop... it bites.
        if name == 'yaml':
            return object.__getattribute__(self, name)
        try:
            return self.fields[name]
        except KeyError:
            return object.__getattribute__(self, name)

    def __init__(self, pod_path, _pod):
        utils.validate_name(pod_path)
        self.pod = _pod
        self.pod_path = pod_path
        self.collection_path = Collection.clean_collection_path(pod_path)
        self.basename = os.path.basename(self.collection_path)
        self.blueprint_path = os.path.join(
            self.pod_path, Collection.BLUEPRINT_PATH)

    def __iter__(self):
        for doc in self.list_docs():
            yield doc

    def __repr__(self):
        return '<Collection "{}">'.format(self.collection_path)

    @classmethod
    def clean_collection_path(cls, pod_path):
        return Collection._content_path_regex.sub('', pod_path).strip('/')

    def _add_localized_docs(self, sorted_docs, pod_path, locale, doc):
        for each_locale in doc.locales:
            if each_locale == doc.default_locale and locale != each_locale:
                continue
            base, ext = os.path.splitext(pod_path)
            localized_file_path = '{}@{}{}'.format(base, each_locale, ext)
            if (locale in [utils.SENTINEL, each_locale]
                    and not self.pod.file_exists(localized_file_path)):
                new_doc = doc.localize(each_locale)
                sorted_docs.insert(new_doc)

    def _get_builtin_field(self, name):
        """Returns a builtin field, which is a field prefixed with a `$`. To be
        backwards compatible with the legacy recommendation, we return the
        field unprefixed with `$` if a prefixed field cannot be found."""
        return self.fields.get('${}'.format(name), self.fields.get(name))

    def _owns_doc_at_path(self, pod_path):
        dir_name = os.path.dirname(pod_path)
        doc_blueprint_path = os.path.join(dir_name, self.BLUEPRINT_PATH)
        if doc_blueprint_path == self.blueprint_path:
            return True
        parts = pod_path.split(os.sep)
        for i, part in enumerate(parts):
            path = os.sep.join(parts[:-i])
            doc_blueprint_path = os.path.join(path, self.BLUEPRINT_PATH)
            if self.pod.file_exists(doc_blueprint_path):
                return doc_blueprint_path == self.blueprint_path
        return False

    @classmethod
    def create(cls, collection_path, fields, pod):
        """Creates a new collection by writing a blueprint."""
        collection = cls.get(collection_path, pod)
        if collection.exists:
            raise CollectionExistsError(
                '{} already exists.'.format(collection))
        fields = utils.dump_yaml(fields)
        pod.write_file(collection.blueprint_path, fields)
        return collection

    @classmethod
    def get(cls, collection_path, _pod):
        """Returns a collection object."""
        return cls(collection_path, _pod)

    @classmethod
    def list(cls, pod):
        items = []
        for root, dirs, _ in pod.walk(cls.CONTENT_PATH + '/'):
            for dir_name in dirs:
                pod_path = os.path.join(root, dir_name)
                pod_path = pod_path.replace(pod.root, '')
                col_path = os.path.join(pod_path, cls.BLUEPRINT_PATH)
                if pod.file_exists(col_path):
                    items.append(pod.get_collection(pod_path))
        return items

    @utils.cached_property
    def default_locale(self):
        locale = None
        if self.localization and 'default_locale' in self.localization:
            locale = self.localization['default_locale']
        return self.pod.normalize_locale(locale)

    @property
    def exists(self):
        """Returns whether the collection exists, as determined by whether
        the collection's blueprint exists."""
        return self.pod.file_exists(self.blueprint_path)

    @utils.cached_property
    def features(self):
        """Collection features."""
        coll_features = features.Features(disabled=[
            self.FEATURE_SEPARATE_ROUTING,
        ])

        # Enable the separate_routing feature if there are routes specified.
        if self.yaml.get('routes') is not None:
            coll_features.enable(self.FEATURE_SEPARATE_ROUTING)

        return coll_features

    @utils.cached_property
    def fields(self):
        fields = untag.Untag.untag(
            self.tagged_fields, params={
                'env': untag.UntagParamRegex(self.pod.env.name),
            })
        return {} if not fields else fields

    @utils.cached_property
    def locales(self):
        localized = ('$localization' in self.fields
                     or 'localization' in self.fields)
        if localized:
            # Disable localization with $localization:~.
            if self.localization is None:
                return []
            if 'locales' in self.localization:
                codes = self.localization['locales'] or []
                return self.pod.normalize_locales(
                    locales.Locale.parse_codes(codes))
        return self.pod.list_locales()

    @property
    def localization(self):
        return self._get_builtin_field('localization')

    @property
    def order(self):
        # Default to maxint so unordered collections go to the end.
        return self.fields.get('$order', sys.maxsize)

    @property
    def parent(self):
        parent_pod_path = os.path.normpath(self.pod_path[:-len(self.basename)])
        if parent_pod_path == Collection.CONTENT_PATH:
            return
        return self.pod.get_collection(parent_pod_path)

    @property
    def path_format(self):
        return self._get_builtin_field('path')

    @property
    def root(self):
        return self._get_builtin_field('root')

    @utils.cached_property
    def routes(self):
        return collection_routes.CollectionRoutes(self.fields.get('routes', {}))

    @property
    def tagged_fields(self):
        return copy.deepcopy(self.yaml)

    @property
    def title(self):
        return self._get_builtin_field('title')

    @property
    def view(self):
        return self._get_builtin_field('view')

    @utils.cached_property
    def yaml(self):
        if not self.exists:
            return {}
        # Use the simple loader to prevent circular loading.
        result = utils.parse_yaml(
            self.pod.read_file(self.blueprint_path), pod=self.pod,
            simple_loader=True)
        if result is None:
            return {}
        return result

    def create_doc(self, basename, fields=utils.SENTINEL, body=utils.SENTINEL):
        """Creates a document within the collection."""
        doc_pod_path = os.path.join(self.pod_path, basename)
        doc = self.get_doc(doc_pod_path)
        doc.write(fields=fields, body=body)
        return doc

    def collections(self):
        """Returns collections contained within this collection. Implemented
        as a function to allow future implementation of arguments."""
        items = []
        for root, dirs, _ in self.pod.walk(self.pod_path):
            if root == self.pod.abs_path(self.pod_path):
                for dir_name in dirs:
                    pod_path = os.path.join(self.pod_path, dir_name)
                    items.append(self.pod.get_collection(pod_path))
        return items

    def delete(self):
        if len(self.list_docs(include_hidden=True)):
            text = 'Collections that are not empty cannot be deleted.'
            raise CollectionNotEmptyError(text)
        self.pod.delete_file(self.blueprint_path)

    def get_doc(self, pod_path, locale=None):
        """Returns a document contained in this collection."""
        pod_path = document.Document.clean_localized_path(pod_path, locale)
        # Some uses are using str(None)
        if locale is not None and locale != 'None':
            locale = locales.Locale.from_alias(self.pod, locale)
            locale.set_alias(self.pod)
            localized_path = document.Document.localize_path(pod_path, str(locale))
            if self.pod.file_exists(localized_path):
                pod_path = localized_path
            elif locale.alias:
                localized_path = document.Document.localize_path(pod_path, locale.alias)
                if self.pod.file_exists(localized_path):
                    pod_path = localized_path
        cached = self.pod.podcache.collection_cache.get_document(
            self, pod_path, locale)
        if cached:
            return cached
        doc = document.Document(
            pod_path, locale=locale, _pod=self.pod, _collection=self)
        self.pod.podcache.collection_cache.add_document(doc)
        return doc

    def list_categories(self):
        return self._get_builtin_field('categories') or []

    def list_docs(self, order_by=None, locale=utils.SENTINEL, reverse=None,
                  include_hidden=False, recursive=True, inject=False,
                  order_by_default=None):
        reverse = False if reverse is None else reverse
        if order_by is None:
            order_by = ('order', 'pod_path')
            order_by_default = order_by_default if order_by_default is not None else 0
        elif isinstance(order_by, str):
            order_by = (order_by, 'pod_path')
        key = operator.attrgetter(*order_by)
        sorted_docs = structures.SortedCollection(
            key=key, default=order_by_default)
        if inject:
            injected_docs = self.pod.inject_preprocessors(collection=self)
            if injected_docs is not None:
                sorted_docs = injected_docs
                self.pod.logger.info(
                    'Injected collection -> {}'.format(self.pod_path))
            return reversed(sorted_docs) if reverse else sorted_docs
        for path in self.pod.list_dir(self.pod_path, recursive=recursive):
            pod_path = os.path.join(self.pod_path, path.lstrip('/'))
            # Document is owned by a different collection, skip it.
            if not self._owns_doc_at_path(pod_path):
                continue
            slug, ext = os.path.splitext(os.path.basename(pod_path))
            if (slug.startswith(self.IGNORE_INITIAL)
                    or ext not in messages.extensions_to_formats
                    or not pod_path):
                continue
            try:
                _, locale_from_path = \
                    document.Document.parse_localized_path(pod_path)
                if locale_from_path:
                    if (locale is not None
                            and locale in [utils.SENTINEL, locale_from_path]):
                        new_doc = self.get_doc(
                            pod_path, locale=locale_from_path)
                        if not include_hidden and new_doc.hidden:
                            continue
                        sorted_docs.insert(new_doc)
                    continue
                doc = self.get_doc(pod_path)
                if not include_hidden and doc.hidden:
                    continue
                if locale in [utils.SENTINEL, None]:
                    sorted_docs.insert(doc)
                if locale is None:
                    continue
                if locale == doc.default_locale:
                    sorted_docs.insert(doc)
                else:
                    self._add_localized_docs(
                        sorted_docs, pod_path, locale, doc)
            except Exception as e:
                logging.error('Error loading doc: {}'.format(pod_path))
                raise
        return reversed(sorted_docs) if reverse else sorted_docs

    # Aliases `collection.docs` to `collection.list_docs`. `collection.docs`
    # should be the public and supported way to retrieve documents from a
    # collection.
    docs = list_docs

    def list_docs_unread(self, locale=utils.SENTINEL, recursive=True, inject=False):
        """Lists the docs without triggering a read from the storage system."""
        docs = []
        if inject:
            injected_docs = self.pod.inject_preprocessors(collection=self)
            if injected_docs is not None:
                docs = injected_docs
                self.pod.logger.info(
                    'Injected collection -> {}'.format(self.pod_path))
            return docs
        for path in self.pod.list_dir(self.pod_path, recursive=recursive):
            pod_path = os.path.join(self.pod_path, path.lstrip('/'))
            # Document is owned by a different collection, skip it.
            if not self._owns_doc_at_path(pod_path):
                continue
            slug, ext = os.path.splitext(os.path.basename(pod_path))
            if (slug.startswith(self.IGNORE_INITIAL)
                    or ext not in messages.extensions_to_formats
                    or not pod_path):
                continue
            try:
                _, locale_from_path = \
                    document.Document.parse_localized_path(pod_path)
                if locale_from_path:
                    if (locale is not None
                            and locale in [utils.SENTINEL, locale_from_path]):
                        docs.append(self.get_doc(
                            pod_path, locale=locale_from_path))
                    continue
                docs.append(self.get_doc(pod_path))
            except Exception:
                logging.error('Error loading doc: {}'.format(pod_path))
                raise
        return docs

    def list_servable_documents(self, include_hidden=False, locales=None,
                                inject=None, doc_list=None):
        if not doc_list:
            inject = False if inject is None else inject
            doc_list = self.list_docs(
                include_hidden=include_hidden, inject=inject)
        docs = []
        for doc in doc_list:
            if (self._get_builtin_field('draft')
                    or not doc.has_serving_path()
                    or not doc.view
                    or (locales and doc.locale not in locales)):
                continue

            # If there is a localized version of the default locale
            # the base document should not be added with the same locale.
            if (doc.locale and doc.locale == doc.default_locale and
                    not document.Document.is_localized_path(doc.pod_path)):
                localized_path = document.Document.localize_path(
                    doc.pod_path, doc.locale)
                if self.pod.file_exists(localized_path):
                    continue

            docs.append(doc)
        return docs

    def list_servable_document_locales(self, pod_path, include_hidden=False,
                                       locales=None):
        sorted_docs = structures.SortedCollection(key=None)
        doc = self.get_doc(pod_path)

        try:
            doc.has_serving_path()  # Using doc fields forces file read.
        except document_front_matter.BadFormatError:
            # Skip adding invalid documents.
            return []

        sorted_docs.insert(doc)
        self._add_localized_docs(sorted_docs, pod_path, utils.SENTINEL, doc)
        return self.list_servable_documents(
            include_hidden=include_hidden, locales=locales, doc_list=sorted_docs)

    def titles(self, title_name=None):
        if title_name is None:
            return self.title
        titles = self.fields.get('$titles', {})
        return titles.get(title_name, self.title)

    def to_message(self):
        message = messages.CollectionMessage()
        message.title = self.title
        message.collection_path = self.collection_path
        return message



"""Subcommand for working with translations."""

import click
from grow.commands.subcommands import translations_diff
from grow.commands.subcommands import translations_download
from grow.commands.subcommands import translations_extract
from grow.commands.subcommands import translations_filter
from grow.commands.subcommands import translations_import
from grow.commands.subcommands import translations_upload


@click.group()
def translations():
    """Translation operations for the pod."""
    pass


# Add the sub commands.
translations.add_command(translations_diff.translations_diff)
translations.add_command(translations_download.translations_download)
translations.add_command(translations_extract.translations_extract)
translations.add_command(translations_filter.translations_filter)
translations.add_command(translations_import.translations_import)
translations.add_command(translations_upload.translations_upload)



"""Response handlers for dev server."""

import logging
import mimetypes
import re
import sys
import traceback
import jinja2
import webob
# NOTE: exc imported directly, webob.exc doesn't work when frozen.
from webob import exc as webob_exc
from werkzeug import wrappers
from werkzeug.exceptions import NotFound
from grow.documents import document
from grow.pods import errors
from grow.pods import ui


class Request(webob.Request):
    pass


class Response(webob.Response):
    default_conditional_response = True


def serve_console(pod, _request, _matched, **_kwargs):
    """Serve the default console page."""
    kwargs = {'pod': pod}
    env = ui.create_jinja_env()
    template = env.get_template('/views/base.html')
    content = template.render(kwargs)
    response = wrappers.Response(content)
    response.headers['Content-Type'] = 'text/html'
    return response


def serve_exception(pod, request, exc, **_kwargs):
    """Serve the exception page."""
    debug = True
    logging.exception
    if isinstance(exc, webob_exc.HTTPException):
        status = exc.status_int
        logging.exception('{}: {}'.format(status, request.path))
    elif isinstance(exc, errors.RouteNotFoundError):
        status = 404
        logging.error('{}: {}'.format(status, request.path))
    elif isinstance(exc, NotFound):
        status = 404
        logging.error('{}: {}'.format(status, request.path))
    else:
        status = 500
        logging.exception('{}: {} - {}'.format(status, request.path, exc))
    env = ui.create_jinja_env()
    template = env.get_template('/views/error.html')
    if (isinstance(exc, errors.BuildError)):
        t_back = exc.traceback
    else:
        unused_error_type, unused_value, t_back = sys.exc_info()
    formatted_traceback = [
        re.sub('^  ', '', line)
        for line in traceback.format_tb(t_back)]
    formatted_traceback = '\n'.join(formatted_traceback)
    kwargs = {
        'exception': exc,
        'is_web_exception': isinstance(exc, webob_exc.HTTPException),
        'pod': pod,
        'status': status,
        'traceback': formatted_traceback,
    }
    if (isinstance(exc, errors.BuildError)):
        kwargs['build_error'] = exc.exception
    if (isinstance(exc, errors.BuildError)
            and isinstance(exc.exception, jinja2.TemplateSyntaxError)):
        kwargs['template_exception'] = exc.exception
    elif isinstance(exc, jinja2.TemplateSyntaxError):
        kwargs['template_exception'] = exc
    content = template.render(**kwargs)
    response = wrappers.Response(content, status=status)
    response.headers['Content-Type'] = 'text/html'
    return response


def serve_pod(pod, request, matched, **_kwargs):
    """Serve pod contents using the routing."""
    controller = pod.router.get_render_controller(
        request.path, matched.value, params=matched.params)
    response = None
    headers = controller.get_http_headers()
    jinja_env = pod.render_pool.get_jinja_env(
        controller.doc.locale) if controller.use_jinja else None
    rendered_document = controller.render(jinja_env=jinja_env, request=request)
    content = rendered_document.read()
    response = Response(body=content)
    response.headers.update(headers)

    if pod.podcache.is_dirty:
        pod.podcache.write()

    return response


def serve_ui_tool(pod, _request, values, **_kwargs):
    tool_path = 'node_modules/{}'.format(values.get('tool'))
    response = wrappers.Response(pod.read_file(tool_path))
    guessed_type = mimetypes.guess_type(tool_path)
    mime_type = guessed_type[0] or 'text/plain'
    response.headers['Content-Type'] = mime_type
    return response


def serve_run_preprocessor(pod, _request, values):
    name = values.get('name')
    if name:
        pod.preprocess([name])
        out = 'Finished preprocessor run -> {}'.format(name)
    else:
        out = 'No preprocessor found.'
    response = wrappers.Response(out)
    response.headers['Content-Type'] = 'text/plain'
    return response



"""Dependency graph for content references."""

import fnmatch
import os
from collections import OrderedDict


class DependencyGraph(object):
    """Dependency graph for tracking relationships between the pod content."""

    def __init__(self):
        self._dependents = {}
        self._dependencies = {}
        self._is_dirty = False

    @staticmethod
    def normalize_path(pod_path):
        """Normalize a pod path."""
        if pod_path and not pod_path.startswith('/'):
            pod_path = '/{}'.format(pod_path)
        return pod_path

    @property
    def is_dirty(self):
        """Have the contents of the dependency graph been modified?"""
        return self._is_dirty

    def add_all(self, path_to_dependencies):
        """Add all from a dict of paths to dependencies."""
        for path, dependencies in path_to_dependencies.items():
            self.add_references(path, dependencies)

    def add(self, source, reference):
        """Add reference made in a source file to the graph."""
        if not source or not reference:
            return

        source = DependencyGraph.normalize_path(source)
        reference = DependencyGraph.normalize_path(reference)

        if source not in self._dependencies:
            self._dependencies[source] = set()
        self._dependencies[source].add(reference)

        # Source are a dependent to themselves.
        if source not in self._dependents:
            self._dependents[source] = set()
        if source not in self._dependents[source]:
            self._is_dirty = True
        self._dependents[source].add(source)

        # Bi-directional dependency references for easier lookup.
        if reference not in self._dependents:
            self._dependents[reference] = set()
        if source not in self._dependents[reference]:
            self._is_dirty = True
        self._dependents[reference].add(source)

    def add_references(self, source, references):
        """Add references made in a source file to the graph."""
        if not references:
            return

        source = DependencyGraph.normalize_path(source)

        self._dependencies[source] = set(references)

        # Source are a dependent to themselves.
        if source not in self._dependents:
            self._dependents[source] = set()
        if source not in self._dependents[source]:
            self._is_dirty = True
        self._dependents[source].add(source)

        # Bi-directional dependency references for easier lookup.
        for reference in references:
            reference = DependencyGraph.normalize_path(reference)
            if reference not in self._dependents:
                self._dependents[reference] = set()

            # Track when the dependency graph has changed.
            if source not in self._dependents[reference]:
                self._is_dirty = True
            self._dependents[reference].add(source)

    def export(self):
        """Formats the dependency graph for export."""
        result = OrderedDict()
        for key in sorted(self._dependencies.keys()):
            result[key] = sorted(list(self._dependencies[key]))
        return result

    def get_dependents(self, reference):
        """
        Gets dependents that rely upon the reference or a collection that
        contains the reference.
        """
        return (self._dependents.get(reference, set())
                | self._dependents.get(os.path.dirname(reference), set())
                | set([reference]))

    def get_dependencies(self, source):
        """Get the dependencies of a specific source."""
        return self._dependencies.get(source, set())

    def mark_clean(self):
        """Mark that the dependency graph is clean."""
        self._is_dirty = False

    def match_dependents(self, reference):
        """
        Match dependents that rely upon the reference or a collection that
        contains the reference using a glob pattern.
        """
        matched_dependents = set()
        for dependent in self._dependents:
            if fnmatch.fnmatch(dependent, reference):
                matched_dependents = (
                    matched_dependents | self._dependents.get(dependent) | set([dependent]))
        return matched_dependents

    def reset(self):
        """Reset all the dependency tracking."""
        self._dependents = {}
        self._dependencies = {}
        self._is_dirty = False



"""Base class for the jinja extension hook."""

from grow.extensions.hooks import base_hook


class JinjaExtensionHook(base_hook.BaseHook):
    """Hook for jinja extension."""

    KEY = 'jinja_extensions'
    NAME = 'Jinja Extensions'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, *_args, **_kwargs):
        """Trigger the jinja extension hook."""
        if previous_result:
            return previous_result
        return []



# -*- coding: utf-8 -*-
"""Footnotes for documents using chicago manual style.

By default specific languages forgo the symbol based footnotes.

https://en.wikipedia.org/wiki/Dagger_(typography) :
While daggers are freely used in English-language texts, they are often avoided
in other languages because of their similarity to the Christian cross. In
German, for example, daggers are commonly employed only to indicate a person's
death or the extinction of a word, language, species or the like.

http://graphicdesign.stackexchange.com/questions/10892/footnote-typographic-conventions :
In Germany the two footnote signs * † have also the meaning of born (*) and died (†)
"""

import collections
import re

SYMBOLS = [
    '*',
    '†',
    '‡',
    '§',
    '||',
    '¶',
    '#',
]
NUMERICAL_SYMBOLS = {
    '0': '⁰',
    '1': '¹',
    '2': '²',
    '3': '³',
    '4': '⁴',
    '5': '⁵',
    '6': '⁶',
    '7': '⁷',
    '8': '⁸',
    '9': '⁹',
}
NUMERIC_LOCALES_REGEX = re.compile(r'_(DE)$', re.IGNORECASE)


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class DuplicateSymbolError(Error, KeyError):
    pass


def symbol_generator(symbols=SYMBOLS):
    loop_count = 1
    loop_index = 0
    while True:
        if loop_index > len(symbols) - 1:
            loop_count += 1
            loop_index = 0
        yield symbols[loop_index] * loop_count
        loop_index += 1


def numberic_symbol_generator():
    index = 1
    while True:
        yield ''.join([NUMERICAL_SYMBOLS[char] for char in str(index)])
        index += 1


def numberic_generator():
    index = 1
    while True:
        yield index
        index += 1


class Footnotes(object):
    def __init__(self, locale, symbols=None, use_numeric=None,
            use_numeric_symbols=None, numeric_locales_pattern=None):
        self.symbol_to_footnote = collections.OrderedDict()
        self.symbols = symbols or SYMBOLS
        self.locale = locale
        self.use_numeric = use_numeric
        self.use_numeric_symbols = use_numeric_symbols
        if numeric_locales_pattern is not None:
            self.numeric_locales_pattern = re.compile(
                numeric_locales_pattern, re.IGNORECASE)
        else:
            self.numeric_locales_pattern = NUMERIC_LOCALES_REGEX
        self.reset()

    def __getitem__(self, key):
        return self.symbol_to_footnote[key]

    def __iter__(self):
        return self.symbol_to_footnote.items()

    def __len__(self):
        return len(self.symbol_to_footnote)

    @property
    def footnotes(self):
        return self.symbol_to_footnote

    @property
    def is_numeric(self):
        return (self.use_numeric
            or self.use_numeric_symbols
            or self.is_numeric_territory)

    @property
    def is_numeric_territory(self):
        if self.locale is None:
            return False
        # When explicitly not using numbers, do not use numeric.
        if self.use_numeric_symbols is False or self.use_numeric is False:
            return False
        return self.numeric_locales_pattern.search(self.locale) is not None

    def add(self, value, custom_symbol=None):
        for symbol, note_value in self.symbol_to_footnote.items():
            if value == note_value:
                return symbol

        if custom_symbol:
            # Check that the symbol is not in use.
            if custom_symbol in self.symbol_to_footnote.keys():
                raise DuplicateSymbolError(
                    'Custom symbol already in use: {}'.format(custom_symbol))
            self.symbol_to_footnote[custom_symbol] = value
            return custom_symbol

        # Doesn't exist, add as new symbol.
        symbol = next(self.generator)
        self.symbol_to_footnote[symbol] = value
        return symbol

    def index(self, key):
        return list(self.symbol_to_footnote.keys()).index(key)

    def items(self):
        return self.symbol_to_footnote.items()

    def reset(self):
        self.symbol_to_footnote = collections.OrderedDict()
        if self.is_numeric:
            if self.use_numeric:
                self.generator = numberic_generator()
            else:
                self.generator = numberic_symbol_generator()
        else:
            self.generator = symbol_generator(self.symbols)



"""Logging utilities for grow."""

import logging

def get_logger(name='pod'):
    """Creates a logger with formatting and a stream handler."""
    handler = logging.StreamHandler()
    _formatter = logging.Formatter('[%(asctime)s] %(message)s', '%H:%M:%S')
    handler.setFormatter(_formatter)
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.addHandler(handler)
    return logger

LOGGER = get_logger('pod')



"""Subcommand for converting pod files from earlier grow versions."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage
from grow.conversion import collection_routing
from grow.conversion import content_locale_split


CFG = rc_config.RC_CONFIG.prefixed('grow.convert')
CONVERT_CHOICES = [
    'content_locale_split',
    'collection_routing',
]


@click.command()
@shared.pod_path_argument
@shared.deployment_option(CFG)
@click.option('--type', 'convert_type', type=click.Choice(CONVERT_CHOICES))
def convert(pod_path, convert_type, deployment):
    """Converts pod files from an earlier version of Grow."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    if deployment:
        deployment_obj = pod.get_deployment(deployment)
        pod.set_env(deployment_obj.config.env)

    if convert_type == 'content_locale_split':
        content_locale_split.Converter.convert(pod)
    elif convert_type == 'collection_routing':
        collection_routing.Converter.convert(pod)
    else:
        raise click.UsageError(
            'Unable to convert files without a --type option.\n'
            'Run `grow convert --help` to see valid --type values.')
    return pod



"""Grow CLI commands."""

from grow.commands.subcommands import build
from grow.commands.subcommands import convert
from grow.commands.subcommands import deploy
from grow.commands.subcommands import inspect
from grow.commands.subcommands import install
from grow.commands.subcommands import preprocess
from grow.commands.subcommands import run
from grow.commands.subcommands import stage
from grow.commands.subcommands import translations


def add_subcommands(group):
    """Add all subcommands to a group."""
    group.add_command(build.build)
    group.add_command(convert.convert)
    group.add_command(deploy.deploy)
    group.add_command(inspect.inspect)
    group.add_command(install.install)
    group.add_command(preprocess.preprocess)
    group.add_command(run.run)
    group.add_command(stage.stage)
    group.add_command(translations.translations)



"""Gulp installer class."""

import subprocess
from grow.sdk.installers import base_installer


class GulpInstaller(base_installer.BaseInstaller):
    """Gulp installer."""

    KIND = 'gulp'

    @property
    def should_run(self):
        """Should the installer run?"""
        return self.pod.file_exists('/gulpfile.js')

    def check_prerequisites(self):
        """Check if required prerequisites are installed or available."""
        status_command = 'gulp --version > /dev/null 2>&1'
        not_found = subprocess.call(
            status_command, **self.subprocess_args(shell=True)) == 127
        if not_found:
            install_commands = [
                'Either add gulp to package.json or install globally using:'
                ' `sudo npm install -g gulp`']
            raise base_installer.MissingPrerequisiteError(
                'The `gulp` command was not found.', install_commands=install_commands)

    def install(self):
        """Install dependencies."""
        pass



"""Subcommand for diffing two directories of translation catalogs."""

import os
import click
from grow.commands import shared
from grow.commands.subcommands.translations_extract import validate_locales
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.diff')


@click.command(name='diff')
@shared.pod_path_argument
@click.option('--translations', '-t', help='Path to the other translations directory to diff.')
@click.option('--out_dir', '-o', help='Where to write the diffs.')
def translations_diff(pod_path, translations, out_dir):
    """Diffs a translations directory with another."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    translations = os.path.join(translations, 'messages.pot')
    other_catalogs = pod.get_catalogs(translations)
    pod.catalogs.diff(other_catalogs, out_dir)
    return pod



"""Nvm installer class."""

import subprocess
from grow.sdk.installers import base_installer


class NvmInstaller(base_installer.BaseInstaller):
    """Nvm installer."""

    KIND = 'nvm'

    @property
    def should_run(self):
        """Should the installer run?"""
        return self.pod.file_exists('/.nvmrc')

    def check_prerequisites(self):
        """Check if required prerequisites are installed or available."""
        status_command = '. $NVM_DIR/nvm.sh && nvm --version > /dev/null 2>&1'
        not_found = subprocess.call(
            status_command, **self.subprocess_args(shell=True)) == 127
        if not_found:
            install_commands = [
                'Download nvm from https://github.com/creationix/nvm']
            raise base_installer.MissingPrerequisiteError(
                'The `nvm` command was not found.', install_commands=install_commands)

    def install(self):
        """Install dependencies."""
        install_command = '. $NVM_DIR/nvm.sh && nvm install'
        process = subprocess.Popen(
            install_command, **self.subprocess_args(shell=True))
        code = process.wait()
        if not code:
            return
        raise base_installer.InstallError(
            'There was an error running `{}`.'.format(install_command))



"""Base extension class for writing new extensions."""

from grow.common import features


class Error(Exception):
    """Base error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class MissingHookError(Error):
    """Missing hook error."""
    pass


class BaseExtension(object):
    """Base extension for custom extensions."""

    def __init__(self, pod, config):
        self.pod = pod
        self.config = config
        self.hooks = features.Features(default_enabled=False)

        if 'enabled' in self.config:
            for hook in self.config['enabled']:
                self.hooks.enable(hook)
        else:
            for hook in self.available_hooks:
                self.hooks.enable(hook.KEY)

        if 'disabled' in self.config:
            for hook in self.config['disabled']:
                self.hooks.disable(hook)

    @property
    def available_hooks(self):
        """Returns the available hook classes."""
        return []

    def auto_hook(self, key):
        """Search for the hook in the available hooks and create."""

        # Allow for defining a *_hook method in the extension.
        hook_method = '{}_hook'.format(key)
        if hasattr(self, hook_method):
            return getattr(self, hook_method)()

        for hook in self.available_hooks:
            if hook.KEY == key:
                return hook(self)
        raise MissingHookError(
            'Hook was not found in extension: {}'.format(key))



"""Dev file change hook."""

from grow.extensions.hooks import base_hook


class DevFileChangeHook(base_hook.BaseHook):
    """Hook for dev file change."""

    KEY = 'dev_file_change'
    NAME = 'Dev File Change'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, pod_path, *_args, **_kwargs):
        """Trigger the pre render hook."""
        return previous_result



from protorpc import messages


class Kind(messages.Enum):
    RENDERED = 1
    STATIC = 2
    SITEMAP = 3


class RouteMessage(messages.Message):
    path = messages.StringField(1)
    kind = messages.EnumField(Kind, 2)
    locale = messages.StringField(3)


class Format(messages.Enum):
    YAML = 1
    MARKDOWN = 2
    HTML = 3


extensions_to_formats = {
    '.html': Format.HTML,
    '.md': Format.MARKDOWN,
    '.yaml': Format.YAML,
}


class Type(messages.Enum):
    STRING = 1
    TEXT = 2
    HTML = 3
    DATETIME = 4
    BOOLEAN = 5
    NUMBER = 6
    COLLECTION = 7


strings_to_types = {
    'String': Type.STRING,
    'Text': Type.TEXT,
    'Html': Type.HTML,
    'DateTime': Type.DATETIME,
    'Boolean': Type.BOOLEAN,
    'Number': Type.NUMBER,
    'Collection': Type.COLLECTION,
}


class FieldMessage(messages.Message):
    key = messages.StringField(1)
    name = messages.StringField(2)
    description = messages.StringField(3)
    type = messages.EnumField(Type, 4)
    repeated = messages.BooleanField(5)
    value = messages.StringField(6)


class CollectionMessage(messages.Message):
    collection_path = messages.StringField(1)  # posts
    num_documents = messages.IntegerField(2)
    title = messages.StringField(3)
    fields = messages.StringField(4)  # JSON fields
#  fields = messages.MessageField(FieldMessage, 2, repeated=True)


class UserMessage(messages.Message):
    name = messages.StringField(1)
    email = messages.StringField(2)


class BuiltInFieldsMessage(messages.Message):
    path = messages.StringField(1)
    view = messages.StringField(2)
    order = messages.IntegerField(3)
    title = messages.StringField(4)


class DocumentMessage(messages.Message):
    doc_path = messages.StringField(1)  # posts/slug.md
    basename = messages.StringField(2)
    collection_path = messages.StringField(3)
    fields = messages.StringField(5)  # JSON fields
    body = messages.StringField(6)
    content = messages.StringField(7)
    html = messages.StringField(8)
    serving_path = messages.StringField(9)
    pod_path = messages.StringField(10)


class LocaleGroupMessage(messages.Message):
    group_name = messages.StringField(1)
    regions = messages.StringField(2, repeated=True)
    languages = messages.StringField(3, repeated=True)


class LocalesMessage(messages.Message):
    groups = messages.MessageField(LocaleGroupMessage, 1, repeated=True)


class MessageMessage(messages.Message):
    msgid = messages.StringField(1)
    msgstr = messages.StringField(2)
    description = messages.StringField(3)
    flags = messages.StringField(4, repeated=True)


class CatalogMessage(messages.Message):
    locale = messages.StringField(1)
    messages = messages.MessageField(MessageMessage, 2, repeated=True)


class CatalogsMessage(messages.Message):
    catalogs = messages.MessageField(CatalogMessage, 1, repeated=True)


class RoutesMessage(messages.Message):
    routes = messages.MessageField(RouteMessage, 1, repeated=True)


class FileMessage(messages.Message):
    pod_path = messages.StringField(1)
    content = messages.StringField(2)
    content_b64 = messages.BytesField(3)
    content_url = messages.StringField(4)
    mimetype = messages.StringField(5)


class PodMessage(messages.Message):
    collections = messages.MessageField(CollectionMessage, 2, repeated=True)
    files = messages.MessageField(FileMessage, 3, repeated=True)
    routes = messages.MessageField(RoutesMessage, 4)


class FileSearchMessage(messages.Message):
    prefix = messages.StringField(1)
    depth = messages.StringField(2)



"""Path formatter for formatting grow paths."""

import datetime
import re
import string
from grow.common import structures
from grow.common import utils


HTML_EXTENSIONS = ('.html', '.htm')
VALID_DOC_EXTENSIONS = ('.html', '.htm', '.xml', '.svg')
INDEX_BASE_ENDINGS = ('/{base}', '/{base}/')


class PathFormat(object):
    """Format url paths using the information from the pod."""

    PARAM_CURLY_REGEX_SECTION = re.compile(r'/{([^}]*)}/')
    PARAM_CURLY_REGEX_END = re.compile(r'/{([^}]*)}$')

    def __init__(self, pod):
        self.pod = pod
        self.formatter = string.Formatter()

    @staticmethod
    def parameterize(path):
        """Replace stubs with routes params."""
        path = re.sub(PathFormat.PARAM_CURLY_REGEX_SECTION, r'/:\1/', path)
        return re.sub(PathFormat.PARAM_CURLY_REGEX_END, r'/:\1', path)

    @staticmethod
    def strip_double_slash(path):
        """Remove double slashes from the path."""
        while '//' in path:
            path = path.replace('//', '/')
        return path

    @staticmethod
    def trailing_slash(doc, path):
        """Adds trailing slash when appropriate."""
        if not doc.view.endswith(HTML_EXTENSIONS):
            return path
        if path.endswith(VALID_DOC_EXTENSIONS):
            return path
        if path and not path.endswith('/'):
            return '{}/'.format(path)
        return path

    @staticmethod
    def _locale_or_alias(locale):
        if not locale:
            return ''
        if not isinstance(locale, str) and locale.alias is not None:
            return locale.alias
        return str(locale)

    def format_doc(self, doc, path, locale=None, parameterize=False):
        """Format a URL path using the doc information."""
        path = '' if path is None else path

        # Most params should always be replaced.
        params = self.params_pod()
        params.update(self.params_doc(path, doc))
        params = self.params_lower(path, params)

        path = utils.safe_format(path, **params)

        if parameterize:
            path = self.parameterize(path)

        params = {}

        if locale is None:
            locale = doc.locale
        params['locale'] = self._locale_or_alias(locale)

        path = utils.safe_format(path, **params)
        path = self.strip_double_slash(path)
        return self.trailing_slash(doc, path)

    def format_pod(self, path, parameterize=False):
        """Format a URL path using the pod information."""
        path = '' if path is None else path

        params = self.params_pod()
        path = utils.safe_format(path, **params)

        if parameterize:
            path = self.parameterize(path)

        return self.strip_double_slash(path)

    def format_static(self, path, locale=None, parameterize=False,
                      fingerprint=None):
        """Format a static document url."""

        params = self.params_pod()
        path = utils.safe_format(path, **params)

        if parameterize:
            path = self.parameterize(path)

        params = {}

        if fingerprint:
            params['fingerprint'] = fingerprint

        params['locale'] = self._locale_or_alias(locale)

        if params:
            path = utils.safe_format(path, **params)

        return self.strip_double_slash(path)

    def format_view(self, doc, path, parameterize=False):
        """Format a URL path using the doc information for views."""
        path = '' if path is None else path

        # Most params should always be replaced.
        params = self.params_pod()
        params.update(self.params_doc(path, doc))
        params = self.params_lower(path, params)

        path = utils.safe_format(path, **params)

        if parameterize:
            path = self.parameterize(path)

        return self.strip_double_slash(path)

    def params_doc(self, path, doc):
        """Selective access to the document properties depending on path."""
        params = {}
        # Remove the base when the path ends with the base and is index.
        if doc.base == 'index' and path.endswith(INDEX_BASE_ENDINGS):
            params['base'] = ''
        else:
            params['base'] = doc.base
        params['collection'] = structures.AttributeDict(
            base_path=doc.collection_base_path,
            sub_path=doc.collection_base_path,
            basename=doc.collection.basename,
            root=doc.collection.root)
        if '{category}' in path:
            params['category'] = doc.category
        if '{parent}' in path:
            params['parent'] = doc.parent if doc.parent else utils.DummyDict()
        if '{slug}' in path:
            params['slug'] = doc.slug
        if '{date}' in path:
            if isinstance(doc.date, datetime.datetime):
                params['date'] = doc.date.date()
            else:
                params['date'] = doc.date
        return params

    def params_lower(self, path, params):
        """Update to support lowercase when in the path."""
        if '|lower' in path:
            for key, value in params.items():
                if isinstance(value, str):
                    params['{}|lower'.format(key)] = value.lower()
        return params

    def params_pod(self):
        params = {}
        podspec = self.pod.podspec.get_config()
        if 'root' in podspec:
            params['root'] = podspec['root']
        else:
            params['root'] = ''
        params['env'] = self.pod.env
        return params



"""Podspec helper."""

from grow.translations import locales


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class PodSpecParseError(Error):
    pass


class PodSpec(object):

    def __init__(self, yaml, pod):
        yaml = yaml or {}
        self.yaml = yaml
        self.pod = pod
        self.grow_version = yaml.get('grow_version')
        _default_locale = yaml.get('localization', {}).get('default_locale', None)
        self.default_locale = locales.Locale.parse(_default_locale)
        self.fields = yaml

    def get_config(self):
        return self.yaml

    def __getattr__(self, name):
        if name in self.fields:
            return self.fields[name]
        tagged_name = '{}@'.format(name)
        if tagged_name in self.fields:
            return self.fields[tagged_name]
        return object.__getattribute__(self, name)

    def __iter__(self):
        return self.yaml.__iter__()

    @property
    def home(self):
        return self.pod.get_home_doc()

    @property
    def root(self):
        return self.fields.get('root', '')

    @property
    def localization(self):
        return self.fields.get('localization')

    def get_locale_alias(self, locale):
        """Get the locale alias for a given locale."""
        if 'localization' in self.yaml and 'aliases' in self.yaml['localization']:
            aliases = self.yaml['localization']['aliases']
            for custom_locale, babel_locale in aliases.items():
                normalized_babel_locale = babel_locale.lower()
                if locale == normalized_babel_locale:
                    return custom_locale
        return locale



"""Local development server manager."""

import logging
import os
import socket
import sys
import threading
from grow.common import bulk_errors
from grow.common import colors
from grow.common import timer
from grow.documents import document
from grow.preprocessors import file_watchers
from grow.server import main as main_lib
from werkzeug import serving


# Number of tries to find a free port.
NUM_TRIES = 10


class CallbackHTTPServer(serving.ThreadedWSGIServer):

    def server_activate(self):
        super(CallbackHTTPServer, self).server_activate()
        _, port = self.server_address
        self.pod.env.port = port
        with timer.Timer() as router_time:
            try:
                self.pod.router.add_all(concrete=False)
            except Exception:
                self.pod.logger.exception('Error encountered starting server.')
                # Do not exit the process here; the server should remain active
                # so the developer can fix any issues related to the router,
                # without needing to manually restart the server.
        self.pod.logger.info('{} routes built in {:.3f} s'.format(
            len(self.pod.router.routes), router_time.secs))

        url, extra_urls = print_server_ready_message(self.pod, self.pod.env.host, port)
        if self.open_browser:
            start_browser_in_thread(url)
            for extra_url in extra_urls:
                start_browser_in_thread(extra_url)

class ServerMessages:
    """Simple server messages."""

    def __init__(self):
        self.messages = []

    def add_message(self, label, message, label_color=None, message_color=None):
        self.messages.append((label, message, label_color, message_color))

    def print(self, log_function):
        label_max_width = 0
        for label, _, _, _ in self.messages:
            label_max_width = max(len(label), label_max_width)

        for label, message, label_color, message_color in self.messages:
            if label_color:
                label = colors.stylize(label.rjust(label_max_width), label_color)
            if message_color:
                message = colors.stylize(message, message_color)
            log_function('{} {}'.format(label, message))


def print_server_ready_message(pod, host, port):
    try:
        home_doc = pod.get_home_doc()
        root_path = home_doc.url.path if home_doc and home_doc.exists else '/'
    except:
        # Allow the user to fix the problem without restarting the server manually.
        root_path = '/'
        pod.logger.exception('Failed to determine root URL path.')
    url_base = 'http://{}:{}/'.format(host, port)
    url_root = 'http://{}:{}{}'.format(host, port, root_path)

    messages = ServerMessages()
    messages.add_message('Pod:', pod.root, colors.HIGHLIGHT)
    messages.add_message('Server:', url_root, colors.HIGHLIGHT)

    # Trigger the dev manager message hook.
    extra_urls = pod.extensions_controller.trigger(
        'dev_manager_message', messages.add_message, url_base, url_root) or []

    messages.add_message(
        'Ready.', 'Press ctrl-c to quit.', colors.SUCCESS, colors.SUCCESS)

    messages.print(pod.logger.info)

    return (url_root, extra_urls)


def start(pod, host=None, port=None, open_browser=False, debug=False,
          preprocess=True, update_check=False):
    main_observer, podspec_observer = file_watchers.create_dev_server_observers(pod)
    if preprocess:
        thread = threading.Thread(target=pod.preprocess, kwargs={'build': False})
        thread.setDaemon(True)
        thread.start()
    port = 8080 if port is None else int(port)
    host = 'localhost' if host is None else host
    patch_broken_pipe_error()
    # Not safe for multi-pod serving env.
    CallbackHTTPServer.pod = pod
    CallbackHTTPServer.open_browser = open_browser
    CallbackHTTPServer.update_check = update_check
    serving.ThreadedWSGIServer = CallbackHTTPServer
    app = main_lib.create_wsgi_app(pod, host, port, debug=debug)
    serving._log = lambda *args, **kwargs: ''
    handler = main_lib.RequestHandler
    num_tries = 0
    done = False

    while num_tries < NUM_TRIES and not done:
        try:
            app.app.port = port
            serving.run_simple(host, port, app, request_handler=handler, threaded=True)
            done = True
        except socket.error as e:
            # if any(x in str(e) for x in ('Errno 48', 'Errno 98')):
            if 'Errno 48' in str(e):
                num_tries += 1
                port += 1
            else:
                # Clean up the file watchers.
                main_observer.stop()
                podspec_observer.stop()
                raise e
        finally:
            if done:
                if pod.podcache.is_dirty:
                    pod.podcache.write()
                # Clean up the file watchers.
                main_observer.stop()
                podspec_observer.stop()
                return

    text = 'Unable to find a port for the server (tried {}).'
    pod.logger.error(text.format(port))
    sys.exit(-1)


def start_browser_in_thread(url):
    def _start_browser():
        import webbrowser
        webbrowser.open(url)
    thread = threading.Thread(target=_start_browser)
    thread.setDaemon(True)
    thread.start()


def patch_broken_pipe_error():
    from socketserver import BaseServer
    from wsgiref import handlers

    handle_error = BaseServer.handle_error
    log_exception = handlers.BaseHandler.log_exception

    def is_broken_pipe_error():
        _, err, _ = sys.exc_info()
        return repr(err) == "error(32, 'Broken pipe')"

    def patched_handle_error(self, request, client_address):
        if not is_broken_pipe_error():
            handle_error(self, request, client_address)

    def patched_log_exception(self, exc_info):
        if not is_broken_pipe_error():
            log_exception(self, exc_info)

    BaseServer.handle_error = patched_handle_error
    handlers.BaseHandler.log_exception = patched_log_exception



"""Threaded loader that forces a list of docs to be loaded from filesystem."""

import sys
import traceback
from multiprocessing.dummy import Pool as ThreadPool
from grow.common import bulk_errors
from grow.documents import document_front_matter


class Error(Exception):
    """Base loading error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class LoadError(Error):
    """Error that occured during the loading."""

    def __init__(self, message, err, err_tb):
        super(LoadError, self).__init__(message)
        self.err = err
        self.err_tb = err_tb


# pylint: disable=too-few-public-methods
class DocsLoader(object):
    """Loader that threads the docs' file system reads."""

    MAX_POOL_SIZE = 100
    MIN_POOL_COUNT = 50
    POOL_RATIO = 0.02

    @staticmethod
    def expand_locales(pod, docs):
        """Expand out the docs list to have the docs for all locales."""
        with pod.profile.timer('DocsLoader.expand_locales'):
            expanded_docs = []
            for doc in docs:
                locales = doc.locales
                if not locales:
                    expanded_docs.append(pod.get_doc(doc.pod_path, None))
                    continue
                for locale in doc.locales:
                    locale_doc = pod.get_doc(doc.pod_path, str(locale))
                    if locale_doc.exists:
                        expanded_docs.append(locale_doc)
                if doc.default_locale not in locales:
                    locale_doc = pod.get_doc(doc.pod_path, doc.default_locale)
                    if locale_doc.exists:
                        expanded_docs.append(locale_doc)
            return expanded_docs

    @staticmethod
    def fix_default_locale(pod, docs, ignore_errors=False):
        """Fixes docs loaded without with the wronge default locale."""
        with pod.profile.timer('DocsLoader.fix_default_locale'):
            root_to_locale = {}
            for doc in docs:
                try:
                    # Ignore the docs that are the same as the default locale.
                    kw_locale = doc._locale_kwarg
                    if kw_locale is not None:
                        continue
                    root_path = doc.root_pod_path
                    current_locale = str(doc.locale)
                    safe_locale = str(doc.locale_safe)
                    if (root_path in root_to_locale
                            and current_locale in root_to_locale[root_path]):
                        continue
                    if safe_locale != current_locale:
                        # The None and safe locale is now invalid in the cache
                        # since the front-matter differs.
                        pod.podcache.collection_cache.remove_document_locale(
                            doc, None)
                        pod.podcache.collection_cache.remove_document_locale(
                            doc, doc.locale_safe)

                        # Get the doc for the correct default locale.
                        clean_doc = doc.localize(current_locale)

                        # Cache default locale (based off front-matter).
                        pod.podcache.collection_cache.add_document_locale(
                            clean_doc, None)

                        # If we have already fixed it, ignore any other locales.
                        if root_path not in root_to_locale:
                            root_to_locale[root_path] = []
                        if current_locale not in root_to_locale[root_path]:
                            root_to_locale[root_path].append(current_locale)
                except document_front_matter.BadFormatError:
                    if not ignore_errors:
                        raise

    @classmethod
    def load(cls, pod, docs, ignore_errors=False, tick=None):
        """Force load the provided docs to read from file system."""
        if not docs:
            return

        def load_func(doc):
            """Force the doc to read the source file."""
            result = LoadResult()
            try:
                # Using doc fields forces file read.
                _ = doc.has_serving_path()
            except document_front_matter.BadFormatError:
                if not ignore_errors:
                    raise
            except Exception as err:  # pylint: disable=broad-except
                _, _, err_tb = sys.exc_info()
                result.errors.append(LoadError(
                    "Error loading {}".format(doc.pod_path),
                    err, err_tb))
            finally:
                if tick:
                    tick()
            return result

        with pod.profile.timer('DocsLoader.load'):
            if ThreadPool is None or len(docs) < cls.MIN_POOL_COUNT:
                for doc in docs:
                    load_func(doc)
                return
            pool_size = min(cls.MAX_POOL_SIZE, len(docs) * cls.POOL_RATIO)
            pool_size = int(round(pool_size))
            thread_pool = ThreadPool(pool_size)
            results = thread_pool.imap_unordered(load_func, docs)
            # Loop results to make sure that the threads are all processed.
            errors = []
            for result in results:
                errors = errors + result.errors
            thread_pool.close()
            thread_pool.join()

            if errors:
                text = 'There were {} errors during doc loading.'
                raise bulk_errors.BulkErrors(text.format(len(errors)), errors)

    @classmethod
    def load_from_routes(cls, pod, routes, **kwargs):
        """Force load the docs from the routes."""

        def _doc_from_routes(routes):
            """Generator for docs from the routes."""
            existing_pod_paths = set()
            docs = []

            for _, value, _ in routes.nodes:
                if value.kind != 'doc':
                    continue
                if 'pod_path' in value.meta:
                    pod_path = value.meta['pod_path']
                    if pod_path not in existing_pod_paths:
                        existing_pod_paths.add(pod_path)
                        docs.append(pod.get_doc(pod_path))
            return docs

        cls.load(pod, _doc_from_routes(routes), **kwargs)


class LoadResult(object):
    """Results from a doc loading."""

    def __init__(self):
        self.errors = []



"""Hook controller for working with hooks from the extensions."""


class HookController(object):
    """Controller for working with pod extension hooks."""

    def __init__(self, pod, key, default_hook):
        self.pod = pod
        self.key = key
        self._hooks = [default_hook]

    def register_extensions(self, extensions):
        """Add new extension hooks to the controller."""
        for extension in extensions:
            if extension.hooks.is_enabled(self.key):
                hook = extension.auto_hook(self.key)
                self._hooks.append(hook)

    def trigger(self, *args, **kwargs):
        """Trigger the hook."""
        result = None
        for hook in self._hooks:
            # Hooks can determine when to be run.
            if hook.should_trigger(result, *args, **kwargs):
                timer = self.pod.profile.timer(
                    hook.__module__ + "." + hook.__class__.__name__)
                with timer:
                    result = hook.trigger(result, *args, **kwargs)
        return result



"""Extension hooks."""

from . import deployment_register_hook
from . import dev_file_change_hook
from . import dev_handler_hook
from . import dev_manager_message_hook
from . import jinja_env_init_hook
from . import jinja_extension_hook
from . import post_render_hook
from . import pre_deploy_hook
from . import pre_render_hook
from . import pre_process_hook
from . import podspec_static_dir_hook
from . import router_add_hook
from . import stub_hook

# pylint: disable=invalid-name
DeploymentRegisterHook = deployment_register_hook.DeploymentRegisterHook
DevFileChangeHook = dev_file_change_hook.DevFileChangeHook
DevHandlerHook = dev_handler_hook.DevHandlerHook
DevManagerMessageHook = dev_manager_message_hook.DevManagerMessageHook
JinjaEnvInitHook = jinja_env_init_hook.JinjaEnvInitHook
JinjaExtensionHook = jinja_extension_hook.JinjaExtensionHook
PodspecStaticDirHook = podspec_static_dir_hook.PodspecStaticDirHook
PostRenderHook = post_render_hook.PostRenderHook
PreDeployHook = pre_deploy_hook.PreDeployHook
PreRenderHook = pre_render_hook.PreRenderHook
PreProcessHook = pre_process_hook.PreProcessHook
PreprocessHook = pre_process_hook.PreProcessHook  # Legacy support.
RouterAddHook = router_add_hook.RouterAddHook
StubHook = stub_hook.StubHook

HOOKS = (
    DeploymentRegisterHook,
    DevFileChangeHook,
    DevHandlerHook,
    DevManagerMessageHook,
    JinjaEnvInitHook,
    JinjaExtensionHook,
    PodspecStaticDirHook,
    PostRenderHook,
    PreDeployHook,
    PreRenderHook,
    PreProcessHook,
    RouterAddHook,
)


def generator_wrapper(pod, hook_name, generator, *args, **kwargs):
    """Wrap the output from a generator to do a simple trigger an extension hook."""
    # Make any changes necessary before deploying.
    for item in generator:
        pod.extensions_controller.trigger(hook_name, item, *args, **kwargs)
        yield item



"""Grow translation importers."""

import io
import collections
import copy
import csv
import errno
import logging
import os
import shutil
import tempfile
import zipfile
from babel.messages import catalog
from babel.messages import pofile


default_external_to_babel_locales = collections.defaultdict(list)
builtin_locales = {
    'en-GB': 'en_GB',
    'es-419': 'es_419',
    'fr-CA': 'fr_CA',
    'iw': 'he_IL',
    'no': 'nb_NO',
    'pt-BR': 'pt_BR',
    'pt-PT': 'pt_PT',
    'zh-CN': 'zh_Hans_CN',
    'zh-HK': 'zh_Hant_HK',
    'zh-TW': 'zh_Hant_TW',
}
for key, value in builtin_locales.items():
    default_external_to_babel_locales[key].append(value)


def _mkdir(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            return
        raise


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class Importer(object):

    def __init__(self, pod, include_obsolete=True, untranslated=False):
        self.pod = pod
        self.include_obsolete = include_obsolete
        self.untranslated = untranslated

    def _validate_path(self, path):
        if not os.path.isfile(path):
            raise Error('Not found: {}'.format(path))

    def import_path(self, path, locale=None):
        if path.endswith('.zip'):
            self._validate_path(path)
            return self.import_zip_file(path)
        elif path.endswith('.po'):
            self._validate_path(path)
            return self.import_file(locale, path)
        elif os.path.isdir(path):
            return self.import_dir(path)
        elif path.endswith('.csv'):
            return self.import_csv_file(path)
        else:
            raise Error('Must import a .zip, .csv, .po file, or directory.')

    def import_csv_file(self, path):
        """Imports a CSV file formatted with locales in the header row and
        translations in the body rows."""
        default_locale = self.pod.podspec.localization.get('default_locale', 'en')
        locales_to_catalogs = {}
        with open(path) as fp:
            reader = csv.DictReader(fp)
            for row in reader:
                if default_locale not in row:
                    text = 'Locale {} not found in {}'.format(default_locale, path)
                    raise Error(text)
                msgid = row[default_locale]
                msgid = msgid
                for locale, translation in row.items():
                    if locale == default_locale:
                        continue
                    translation = translation
                    message = catalog.Message(msgid, translation)
                    if locale not in locales_to_catalogs:
                        locales_to_catalogs[locale] = catalog.Catalog()
                    locales_to_catalogs[locale][msgid] = message
        for locale, catalog_obj in locales_to_catalogs.items():
            fp = io.BytesIO()
            pofile.write_po(fp, catalog_obj)
            fp.seek(0)
            content = fp.read()
            self.import_content(locale, content)
            fp.close()

    def import_zip_file(self, zip_path):
        try:
            temp_dir_path = tempfile.mkdtemp()
            with zipfile.ZipFile(zip_path, 'r') as zip_file:
                zip_file.extractall(temp_dir_path)
            return self.import_dir(temp_dir_path)
        finally:
            shutil.rmtree(temp_dir_path)

    def import_dir(self, dir_path):
        # Track if the imported content is actually changing the translations.
        has_changed_content = False

        # Detect whether importing from another pod.
        podspec_path = os.path.join(dir_path, 'podspec.yaml')
        if os.path.exists(podspec_path):
            text = 'Importing from pod -> {}'.format(podspec_path)
            self.pod.logger.info(text)
            translations_dir = os.path.join(dir_path, 'translations')
            for locale in os.listdir(translations_dir):
                po_path = os.path.join(
                        translations_dir, locale, 'LC_MESSAGES', 'messages.po')
                if os.path.exists(po_path):
                    if self.import_file(locale, po_path):
                        has_changed_content = True
            return has_changed_content

        # TODO(jeremydw): Allow a custom syntax for translation importers.
        # Currently, assume one directory per locale.
        for locale in os.listdir(dir_path):
            locale_dir = os.path.join(dir_path, locale)
            if locale.startswith('.') or os.path.isfile(locale_dir):
                continue
            for basename in os.listdir(locale_dir):
                po_path = os.path.join(locale_dir, basename)
                if basename.endswith('.po'):
                    if self.import_file(locale, po_path):
                        has_changed_content = True
                else:
                    self.pod.logger.warning('Skipping: {}'.format(po_path))

        return has_changed_content

    def import_file(self, locale, po_path):
        if not os.path.exists(po_path):
            raise Error('Couldn\'t find PO file: {}'.format(po_path))
        content = open(po_path).read()
        return self.import_content(locale, content)

    def import_content(self, locale, content):
        if locale is None:
            raise Error('Must specify locale.')

        # Track if the imported content is actually changing the translations.
        has_changed_content = False
        imported_translations = 0
        total_translations = 0

        # Leverage user-defined locale identifiers when importing translations.
        external_to_babel_locales = copy.deepcopy(
            default_external_to_babel_locales)
        if self.pod.podspec.localization:
            if 'import_as' in self.pod.podspec.localization:
                import_as = self.pod.podspec.localization['import_as']
                for external_locale, babel_locales in import_as.items():
                    for babel_locale in babel_locales:
                        external_to_babel_locales[external_locale].append(
                            babel_locale)

        babel_locales = external_to_babel_locales.get(locale, [locale])
        for babel_locale in babel_locales:
            pod_translations_dir = os.path.join(
                'translations', babel_locale, 'LC_MESSAGES')
            pod_po_path = os.path.join(pod_translations_dir, 'messages.po')
            if self.pod.file_exists(pod_po_path):
                try:
                    existing_po_file = self.pod.open_file(pod_po_path)
                    existing_catalog = pofile.read_po(
                        existing_po_file, babel_locale)
                    po_file_to_merge = io.StringIO()
                    try:
                        po_file_to_merge.write(content)
                    except TypeError:
                        po_file_to_merge.write(content.decode())
                    po_file_to_merge.seek(0)
                    catalog_to_merge = pofile.read_po(
                        po_file_to_merge, babel_locale)
                except Exception as e:
                    text = 'Error merging catalogs for locale -> {}'
                    logging.error(text.format(locale))
                    raise
                for message in catalog_to_merge:
                    if message.id not in existing_catalog:
                        # Skip messages that don't exist if we are not
                        # including obsolete messages.
                        if not self.include_obsolete:
                            continue
                        existing_catalog[message.id] = message
                        if message.id:  # Only count the non-header messages.
                            imported_translations += 1
                    elif (message.string
                          and existing_catalog[message.id].string != message.string):
                        # Skip messages that have translations in the catalog
                        # if we are only importing untranslated strings.
                        if self.untranslated and existing_catalog[message.id].string:
                            continue
                        # Avoid overwriting with empty/identical strings.
                        existing_catalog[message.id].string = message.string
                        imported_translations += 1

                if imported_translations > 0:
                    has_changed_content = True
                    existing_po_file = self.pod.open_file(pod_po_path, mode='wb')
                    pofile.write_po(existing_po_file, existing_catalog, width=80,
                                    sort_output=True, sort_by_file=True)
                    total_translations = len(catalog_to_merge)
            else:
                # Skip new catalogs if not including obsolete messages.
                if not self.include_obsolete:
                    continue
                has_changed_content = True
                self.pod.write_file(pod_po_path, content)
                existing_po_file = self.pod.open_file(pod_po_path)
                existing_catalog = pofile.read_po(
                    existing_po_file, babel_locale)
                imported_translations = len(existing_catalog)
                total_translations = imported_translations

        return has_changed_content, imported_translations, total_translations



"""Subcommand to run the preprocessors for the pod."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.preprocess')


@click.command()
@shared.pod_path_argument
@click.option('--all', '-A', 'run_all', is_flag=True, default=False,
              help='Whether to run all preprocessors, even if a preprocessor'
                   ' has autorun disabled.')
@click.option('--preprocessor', '-p', type=str, multiple=True,
              help='Name of preprocessor to run. Preprocessors controlled by'
                   ' the preprocess command must have names or tags.')
@click.option('--tag', '-t', type=str, multiple=True,
              help='Tags of preprocessors to run. Preprocessors controlled by'
                   ' the preprocess command must have names or tags.')
@click.option('--ratelimit', type=int,
              help='Limit the execution speed of preprocessors. Grow will '
                   'sleep for X seconds between running each preprocessor, '
                   'where X is the value of --ratelimit. This flag can be '
                   'useful when using mulitple Google APIs-based '
                   'preprocessors on the same resource to avoid rate limit '
                   'errors.')
@shared.deployment_option(CFG)
def preprocess(pod_path, preprocessor, run_all, tag, ratelimit, deployment):
    """Runs preprocessors."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    if deployment:
        deployment_obj = pod.get_deployment(deployment)
        pod.set_env(deployment_obj.config.env)
    with pod.profile.timer('grow_preprocess'):
        pod.preprocess(preprocessor, run_all=run_all, tags=tag, ratelimit=ratelimit)
    return pod



from babel import localedata
from grow.pods import errors
from grow.pods import messages
import pickle
import os
import babel
import re


class Locales(object):

    def __init__(self, pod):
        self.pod = pod

    def list_groups(self):
        if 'locales' not in self.pod.yaml:
            return []
        return list(self.pod.yaml['locales'].keys())

    def get_regions(self, group_name='default'):
        if 'regions' not in self.pod.yaml:
            return []
        try:
            return self.pod.yaml['locales'][group_name].get('regions', [])
        except errors.PodConfigurationError:
            return []

    def get_languages(self, group_name='default'):
        if 'locales' not in self.pod.yaml:
            return []
        try:
            return self.pod.yaml['locales'][group_name].get('languages', [])
        except errors.PodConfigurationError:
            return []

    def to_message(self):
        message = messages.LocalesMessage()
        message.groups = []
        for group_name in self.list_groups():
            group_message = messages.LocaleGroupMessage()
            group_message.group_name = group_name
            group_message.regions = self.get_regions(group_name)
            group_message.languages = self.get_languages(group_name)
            message.groups.append(group_message)
        return message


class Locale(babel.Locale):
    RTL_REGEX = re.compile('^(he|ar|fa|ur)(\W|$)')
    _alias = None

    def __init__(self, language, *args, **kwargs):
        # Normalize from "de_de" to "de_DE" for case-sensitive filesystems.
        parts = language.rsplit('_', 1)
        if len(parts) > 1:
            language = '{}_{}'.format(parts[0], parts[1].upper())
        super(Locale, self).__init__(language, *args, **kwargs)

    @classmethod
    def parse(cls, *args, **kwargs):
        locale = super(Locale, cls).parse(*args, **kwargs)
        # Weak attempt to permit fuzzy locales (locales for which we still have
        # language and country information, but not a full localedata file for),
        # but disallow completely invalid locales. See note at end of file.
        if locale and locale.get_display_name() is None:
            raise ValueError('{} is not a valid locale identifier'.format(args[0]))
        return locale

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self).lower() == other.lower()
        return super(Locale, self).__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '<Locale: "{}">'.format(str(self))

    @classmethod
    def parse_codes(cls, codes):
        return [cls.parse(code) for code in codes]

    @property
    def is_rtl(self):
        return bool(Locale.RTL_REGEX.match(self.language))

    @property
    def direction(self):
        return 'rtl' if self.is_rtl else 'ltr'

    @classmethod
    def from_alias(cls, pod, alias):
        podspec = pod.get_podspec()
        config = podspec.get_config()
        if 'localization' in config and 'aliases' in config['localization']:
            aliases = config['localization']['aliases']
            for custom_locale, babel_locale in aliases.items():
                if custom_locale == alias:
                    return cls.parse(babel_locale)
        return cls.parse(alias)

    def set_alias(self, pod):
        podspec = pod.get_podspec()
        self.alias = podspec.get_locale_alias(str(self).lower())

    @property
    def alias(self):
        return self._alias

    @alias.setter
    def alias(self, alias):
        self._alias = alias


# NOTE: Babel does not support "fuzzy" locales. A locale is considered "fuzzy"
# when a corresponding "localedata" file that matches a given locale's full
# identifier (e.g. "en_US") does not exist. Here's one example: "en_BD". CLDR
# does not have a localedata file matching "en_BD" (English in Bangladesh), but
# it does have individual files for "en" and also "bn_BD". As it turns
# out, localedata files that correspond to a locale's full identifier (e.g.
# "bn_BD.dat") are actually pretty light on the content (largely containing
# things like start-of-week information) and most of the "meat" of the data is
# contained in the main localedata file, e.g. "en.dat".
#
# Users may need to generate pages corresponding to locales that we don't
# have full localedata for, and until Babel supports fuzzy locales, we'll
# monkeypatch two Babel functions to provide partial support for fuzzy locales.
#
# With this monkeypatch, locales will be valid even if Babel doesn't have a
# localedata file matching a locale's full identifier, but locales will still
# fail with a ValueError if the user specifies a territory that does not exist.
# With this patch, a user can, however, specify an invalid language. Obviously,
# this patch should be removed when/if Babel adds support for fuzzy locales.
# Optionally, we may want to provide users with more control over whether a
# locale is valid or invalid, but we can revisit that later.

# See: https://github.com/grow/grow/issues/93


def fuzzy_load(name, merge_inherited=True):
    localedata._cache_lock.acquire()
    try:
        data = localedata._cache.get(name)
        if not data:
            # Load inherited data
            if name == 'root' or not merge_inherited:
                data = {}
            else:
                parts = name.split('_')
                if len(parts) == 1:
                    parent = 'root'
                else:
                    parent = '_'.join(parts[:-1])
                data = fuzzy_load(parent).copy()
            filename = os.path.join(localedata._dirname, '%s.dat' % name)
            try:
                fileobj = open(filename, 'rb')
                try:
                    if name != 'root' and merge_inherited:
                        localedata.merge(data, pickle.load(fileobj))
                    else:
                        data = pickle.load(fileobj)
                    localedata._cache[name] = data
                finally:
                    fileobj.close()
            except IOError:
                pass
        return data
    finally:
        localedata._cache_lock.release()


localedata.exists = lambda name: True
localedata.load = fuzzy_load



"""
Cache for storing and retrieving routing information in a pod.
"""

from grow.routing import router


FILE_ROUTES_CACHE = 'routescache.json'


class RoutesCache(object):
    """Routes cache for caching routing data in a pod."""

    KEY_CONCRETE = 'concrete'
    KEY_DYNAMIC = 'dynamic'
    KEY_NONE = '__None__'

    def __init__(self):
        self._cache = {
            self.KEY_CONCRETE: {},
            self.KEY_DYNAMIC: {},
        }
        self._is_dirty = False

    @classmethod
    def _cache_key(cls, concrete):
        return cls.KEY_CONCRETE if concrete else cls.KEY_DYNAMIC

    @staticmethod
    def _export_cache(routes):
        routing_info = {}
        for env, env_routes in routes.items():
            routing_info[env] = {}
            for path, item in env_routes.items():
                if hasattr(item['value'], 'export'):
                    value = item['value'].export()
                else:
                    value = item['value']
                routing_info[env][path] = {
                    'value': value,
                    'options': item['options'],
                }
        return routing_info

    def add(self, key, value, options=None, concrete=False, env=None):
        """Add a new item to the cache or overwrite an existing value."""
        if env is None:
            env = self.KEY_NONE
        cache_key = self._cache_key(concrete)
        if env not in self._cache[cache_key]:
            self._cache[cache_key][env] = {}
        cache = self._cache[cache_key][env]
        cache_value = {
            'value': value,
            'options': options,
        }
        if not self._is_dirty and (key not in cache or cache[key] != cache_value):
            self._is_dirty = True
        cache[key] = cache_value

    def export(self, concrete=None):
        """Returns the raw cache data."""
        if concrete is None:
            return {
                'version': 1,
                self.KEY_CONCRETE: self._export_cache(self._cache[self.KEY_CONCRETE]),
                self.KEY_DYNAMIC: self._export_cache(self._cache[self.KEY_DYNAMIC]),
            }
        return self._export_cache(self._cache[self._cache_key(concrete)])

    def from_data(self, data):
        """Set the cache from data."""
        # Check for version changes in the data format.
        version = data.get('version')
        if not version or version < 1:
            return

        for super_key in [self.KEY_DYNAMIC, self.KEY_CONCRETE]:
            if super_key in data:
                concrete = super_key == self.KEY_CONCRETE
                for env, env_data in data[super_key].items():
                    for key, item in env_data.items():
                        self.add(
                            key,
                            router.RouteInfo.from_data(**item['value']),
                            options=item['options'], concrete=concrete,
                            env=env)

    def get(self, key, concrete=False, env=None):
        """Retrieve the value from the cache."""
        if env is None:
            env = self.KEY_NONE
        return self._cache[self._cache_key(concrete)].get(env, {}).get(key, None)

    @property
    def is_dirty(self):
        """Have the contents of the object cache been modified?"""
        return self._is_dirty

    def mark_clean(self):
        """Mark that the object cache is clean."""
        self._is_dirty = False

    def raw(self, concrete=None, env=None):
        """Returns the raw cache data."""
        if concrete is None:
            return self._cache
        if env is None:
            env = self.KEY_NONE
        return self._cache[self._cache_key(concrete)].get(env, {})

    def remove(self, key, concrete=False, env=None):
        """Removes a single element from the cache."""
        if env is None:
            env = self.KEY_NONE
        self._is_dirty = True
        return self._cache[self._cache_key(concrete)].get(env, {}).pop(key, None)

    def reset(self):
        """Reset the internal cache object."""
        self._cache = {
            self.KEY_CONCRETE: {},
            self.KEY_DYNAMIC: {},
        }
        self._is_dirty = False



"""Common OAuth functionality."""

import logging
import os

from grow.common import utils
from oauth2client import client
from oauth2client import file as oauth_file
from oauth2client import service_account
from oauth2client import tools

# Silence "Loading" messages from keyring.
# Even though we are not using keyring it is part of the oauth library.
KEYRING_LOG = logging.getLogger('keyring.backend')
KEYRING_LOG.setLevel(logging.WARNING)

logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)


# Google API details for a native/installed application for API project grow-prod.
CLIENT_ID = '578372381550-jfl3hdlf1q5rgib94pqsctv1kgkflu1a.apps.googleusercontent.com'
CLIENT_SECRET = 'XQKqbwTg88XVpaBNRcm_tYLf'  # Not so secret for installed apps.
REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'
DEFAULT_STORAGE_KEY = 'Grow SDK'
BROWSER_API_KEY = 'AIzaSyDCb_WtWJnlLPdL8IGLvcVhXAjaBHbRY5E'

_LAST_LOGGED_EMAIL = None
_CLEARED_AUTH_KEYS = {}

DEFAULT_AUTH_KEY_FILE = 'auth-key.json'


def get_storage(key, username):
    """Returns the Storage class compatible with the current environment."""
    key = utils.slugify(key)
    file_name = os.path.expanduser('~/.config/grow/{}_{}'.format(key, username))
    dir_name = os.path.dirname(file_name)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    return oauth_file.Storage(file_name)


def get_credentials_and_storage(scope, storage_key=DEFAULT_STORAGE_KEY):
    if os.getenv('CLEAR_AUTH') and storage_key not in _CLEARED_AUTH_KEYS:
        clear_credentials(storage_key=storage_key)
    username = os.getenv('AUTH_EMAIL_ADDRESS', 'default')
    storage = get_storage(storage_key, username)
    if 'CI' in os.environ:  # Avoid using keyring in CI/Travis.
        return None, storage
    credentials = storage.get()
    return credentials, storage


def get_or_create_credentials(scope, storage_key=DEFAULT_STORAGE_KEY):
    key_file = os.getenv('AUTH_KEY_FILE')
    # If AUTH_KEY_FILE is unset, use default auth key file if it exists.
    if not key_file and os.path.exists(DEFAULT_AUTH_KEY_FILE):
        key_file = DEFAULT_AUTH_KEY_FILE
    if key_file:
        key_file = os.path.expanduser(key_file)
        return (service_account.
            ServiceAccountCredentials.from_json_keyfile_name(key_file, scope))
    credentials, storage = get_credentials_and_storage(scope,
        storage_key=storage_key)
    if credentials is None:
        parser = tools.argparser
        if os.getenv('INTERACTIVE_AUTH'):
            args = []
        else:
            args = ['--noauth_local_webserver']
        flags, _ = parser.parse_known_args(args)
        flow = client.OAuth2WebServerFlow(CLIENT_ID, CLIENT_SECRET, scope,
                                          redirect_uri=REDIRECT_URI)
        credentials = tools.run_flow(flow, storage, flags)
        # run_flow changes the logging level, so change it back.
        logging.getLogger().setLevel(getattr(logging, 'INFO'))
    # Avoid logspam by logging the email address only once.
    if hasattr(credentials, 'id_token') and credentials.id_token:
        email = credentials.id_token.get('email')
        global _LAST_LOGGED_EMAIL
        if email and _LAST_LOGGED_EMAIL != email:
            logging.info('Authorizing using -> {}'.format(email))
            _LAST_LOGGED_EMAIL = email
    return credentials


def clear_credentials(storage_key=DEFAULT_STORAGE_KEY):
    username = os.getenv('AUTH_EMAIL_ADDRESS', 'default')
    storage = get_storage(storage_key, username)
    storage.delete()
    _CLEARED_AUTH_KEYS[storage_key] = True



"""Cache for storing and retrieving data specific to collections.

Supports caching the collection and caching documents within those collections.
"""

import os
import threading
from grow.collections import collection


class CollectionCache(object):

    def __init__(self):
        self.reset()
        self._lock = threading.RLock()

    @staticmethod
    def generate_cache_key(pod_path, locale):
        return '{}::{}'.format(pod_path, locale)

    def add_collection(self, col):
        self._cache[col.collection_path] = {
            'collection': col,
            'docs': {},
        }

    def add_document(self, doc):
        col = doc.collection
        self.ensure_collection(col)
        # NOTE: Using `doc.locale` causes infinite loop since it needs to load
        # the fields (which can have circular dependencies).
        cache_key = CollectionCache.generate_cache_key(
            doc.pod_path, doc.locale_safe)
        self._cache[col.collection_path]['docs'][cache_key] = doc

    def add_document_locale(self, doc, locale):
        """Force a doc to be saved to a specific locale.

        When docs have a default locale that differs from the collection default
        it causes issues since the cache does not know that None locale is not correct.
        This allows for the None locale to be forced to the default locale from the
        doc.
        """
        col = doc.collection
        self.ensure_collection(col)
        cache_key = CollectionCache.generate_cache_key(
            doc.pod_path, locale)
        self._cache[col.collection_path]['docs'][cache_key] = doc

    def remove_by_path(self, path):
        """Removes the collection or document based on the path."""
        if path.startswith(collection.Collection.CONTENT_PATH):
            if path.endswith(
                    '/{}'.format(collection.Collection.BLUEPRINT_PATH)):
                # If this is a blueprint then remove the entire collection.
                col_path = path[len(collection.Collection.CONTENT_PATH):]
                # Get just the directory.
                col_path = os.path.split(col_path)[0]
                collection_path = col_path[1:]  # Remove /
                with self._lock:
                    if collection_path in self._cache:
                        del self._cache[collection_path]
            else:
                # Search for an existing collection path.
                col_path = path[len(collection.Collection.CONTENT_PATH):]
                col_path = os.path.split(col_path)[0]
                while col_path != os.sep:
                    collection_path = col_path[1:]
                    with self._lock:
                        if collection_path in self._cache:
                            # Do a 'wildcard' match on the path to remove all
                            # locales.
                            generic_key = CollectionCache.generate_cache_key(
                                path, '')
                            for key in self._cache[collection_path]['docs'].keys():
                                if key.startswith(generic_key):
                                    del self._cache[
                                        collection_path]['docs'][key]
                                    return
                    col_path = os.path.split(col_path)[0]

    def remove_collection(self, col):
        with self._lock:
            if col.collection_path in self._cache:
                del self._cache[col.collection_path]

    def remove_document(self, doc):
        col = doc.collection
        with self._lock:
            if col.collection_path in self._cache:
                cache_key = CollectionCache.generate_cache_key(
                    doc.pod_path, doc.locale_safe)
                if cache_key in self._cache[col.collection_path]['docs']:
                    del self._cache[col.collection_path]['docs'][cache_key]

    def remove_document_locale(self, doc, locale):
        col = doc.collection
        with self._lock:
            if col.collection_path in self._cache:
                cache_key = CollectionCache.generate_cache_key(
                    doc.pod_path, locale)
                if cache_key in self._cache[col.collection_path]['docs']:
                    del self._cache[col.collection_path]['docs'][cache_key]

    def remove_document_locales(self, doc):
        col = doc.collection
        with self._lock:
            if col.collection_path in self._cache:
                doc_cache_key = CollectionCache.generate_cache_key(
                    doc.pod_path, '')
                invalid_keys = []
                for cache_key in self._cache[col.collection_path]['docs'].keys():
                    if cache_key.startswith(doc_cache_key):
                        invalid_keys.append(cache_key)
                for cache_key in invalid_keys:
                    del self._cache[col.collection_path]['docs'][cache_key]

    def ensure_collection(self, col):
        with self._lock:
            if col.collection_path not in self._cache:
                self.add_collection(col)

    def get_collection(self, collection_path):
        collection_path = collection.Collection.clean_collection_path(
            collection_path)
        if collection_path in self._cache:
            return self._cache[collection_path]['collection']
        return None

    def get_document(self, col, pod_path, locale):
        if col.collection_path in self._cache:
            cache_key = CollectionCache.generate_cache_key(pod_path, locale)
            return self._cache[col.collection_path]['docs'].get(
                cache_key, None)
        return None

    def reset(self):
        self._cache = {}



"""Template jinja tests."""

def is_subset_of(value, subset):
    """Check if a variable is a subset."""
    return set(value) >= set(subset)

def is_superset_of(value, superset):
    """Check if a variable is a superset."""
    return set(value) <= set(superset)

def create_builtin_tests():
    """Tests standard for the template rendering."""
    return (
        ('subset', is_subset_of),
        ('superset', is_superset_of),
    )



class BaseStorage(object):

    is_cloud_storage = False

    @staticmethod
    def read(filename):
        raise NotImplementedError

    @staticmethod
    def open(filename, mode='r'):
        raise NotImplementedError

    @staticmethod
    def modified(filename):
        raise NotImplementedError

    @staticmethod
    def size(filename):
        raise NotImplementedError

    @staticmethod
    def stat(filename):
        raise NotImplementedError

    @staticmethod
    def listdir(dirpath, recursive=True):
        raise NotImplementedError

    @staticmethod
    def walk(dirpath):
        raise NotImplementedError

    @staticmethod
    def delete(filename):
        raise NotImplementedError

    @staticmethod
    def delete_dir(dirpath):
        raise NotImplementedError

    @staticmethod
    def delete_files(dirpaths, recursive=False, pattern=None):
        raise NotImplementedError

    @staticmethod
    def hash(filename):
        raise NotImplementedError

    @staticmethod
    def JinjaLoader(path):
        raise NotImplementedError

    @staticmethod
    def copy_to(path, target_path):
        raise NotImplementedError

    @staticmethod
    def move_to(path, target_path):
        raise NotImplementedError

    @classmethod
    def write(cls, path, content):
        file_obj = cls.open(path, mode='w')
        file_obj.write(content)
        return file_obj

    @classmethod
    def update_headers(cls, headers, path):
        pass



"""
Cache for storing and retrieving data specific to a document.

Supports caching specific to the pod_path of a document.

The contents of the cache should be raw and not internationalized as it will
be shared between locales with the same pod_path.
"""

class DocumentCache(object):

    def __init__(self):
        self.reset()

    def _ensure_exists(self, doc, value=None):
        if doc.pod_path not in self._cache:
            self._cache[doc.pod_path] = value or {}
        return doc.pod_path

    def add(self, doc, value):
        self._cache[doc.pod_path] = value

    def add_all(self, path_to_cached):
        for path, value in path_to_cached.items():
            self._cache[path] = value

    def add_property(self, doc, prop, value):
        path = self._ensure_exists(doc)
        self._cache[path][prop] = value

    def remove(self, doc):
        return self.remove_by_path(doc.pod_path)

    def remove_by_path(self, path):
        return self._cache.pop(path, None)

    def export(self):
        return self._cache

    def get(self, doc):
        return self._cache.get(doc.pod_path, None)

    def get_property(self, doc, prop):
        if doc.pod_path in self._cache:
            return self._cache[doc.pod_path].get(prop, None)
        return None

    def reset(self):
        self._cache = {}



"""Code timing for profiling."""

import time


class Timer(object):
    """Times code to see how long it takes using a context manager."""

    def __init__(self, key, label=None, meta=None):
        self._time = time
        self.key = key
        self.label = label or key
        self.meta = meta
        self.start = None
        self.end = None

    def __enter__(self):
        return self.start_timer()

    def __exit__(self, *args):
        self.stop_timer()

    def __repr__(self):
        if self.label != self.key:
            return '<Timer:{} {} : {}>'.format(self.key, self.label, self.duration)
        return '<Timer:{} : {}>'.format(self.key, self.duration)

    @property
    def duration(self):
        """Duration of timer."""
        return self.end - self.start

    def export(self):
        """Export the timer data."""
        return {
            'key': self.key,
            'label': self.label,
            'meta': self.meta,
            'start': self.start,
            'end': self.end,
        }

    def start_timer(self):
        """Starts the timer."""
        self.start = self._time.time()
        return self

    def stop_timer(self):
        """Stops the timer."""
        self.end = self._time.time()
        return self


class Profile(object):
    """Keeps track of all of the timer usage."""

    def __init__(self):
        self.timers = []

    def __iter__(self):
        for timer in self.timers:
            yield timer

    def __len__(self):
        return len(self.timers)

    def add_timer(self, timer):
        """Adds a new timer."""
        if timer is None:
            return
        self.timers.append(timer)
        return timer

    def timer(self, *args, **kwargs):
        """Create a new timer."""
        timer = Timer(*args, **kwargs)
        self.timers.append(timer)
        return timer

    def export(self):
        """Export the timer data for each timer created."""
        return [t.export() for t in self.timers]



"""Subcommand for extracting messages for translation."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.translations import catalog_holder
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.extract')


@click.command(name='extract')
@shared.pod_path_argument
@click.option('--init/--no-init', default=False, is_flag=True,
              help='Whether to create an initial set of empty translation'
                   ' catalogs using the locales configured in podpsec.')
@click.option('--update/--no-update', default=True, is_flag=True,
              help='Whether to update translation catalogs with extracted'
                   ' messages. If false, only a catalog template will be'
                   ' created.')
@click.option('--fuzzy-matching/--no-fuzzy-matching',
              default=CFG.get('fuzzy-matching', None), is_flag=True,
              help='Whether to use fuzzy matching when updating translation'
                   ' catalogs. If --fuzzy-matching is specified, updated'
                   ' catalogs will contain fuzzy-translated messages with the'
                   ' "fuzzy" flag. If --fuzzy-matching is not specified,'
                   ' updated catalogs will contain new messages without fuzzy'
                   ' translations. This flag is set to "false" by default'
                   ' because some translation tools may not remove the "fuzzy"'
                   ' flag from messages even after a translation has been'
                   ' provided.')
@click.option('--audit', default=False, is_flag=True,
              help='Audit content files for all untagged strings instead of'
                   ' actually extracting.')
@click.option('-o', type=str, default=None,
              help='Where to write the extracted translation catalog. The path'
                   ' must be relative to the pod\'s root.')
@shared.include_header_option(CFG)
@shared.include_obsolete_option(CFG)
@shared.locale_option(
    help_text='Which locale(s) to analyze when creating template catalogs'
              ' that contain only untranslated messages. This option is'
              ' only applicable when using --update or --init.')
@shared.localized_option(CFG)
@shared.path_option
def translations_extract(pod_path, init, update, include_obsolete, localized,
                         include_header, locale, fuzzy_matching, audit, path, o):
    """Extracts tagged messages from source files into a template catalog."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    with pod.profile.timer('grow_translations_extract'):
        include_obsolete, localized, include_header, use_fuzzy_matching, = \
            pod.catalogs.get_extract_config(
                include_header=include_header, include_obsolete=include_obsolete,
                localized=localized, use_fuzzy_matching=fuzzy_matching)
        locales = validate_locales(pod.list_locales(), locale)
        catalogs = pod.get_catalogs(template_path=o)
        untagged_strings, extracted_catalogs = \
            catalogs.extract(include_obsolete=include_obsolete, localized=localized,
                             include_header=include_header,
                             use_fuzzy_matching=fuzzy_matching, locales=locales,
                             audit=audit, paths=path, out_path=o)
        if audit:
            tables = catalog_holder.Catalogs.format_audit(
                untagged_strings, extracted_catalogs)
            # NOTE: Should use click.echo_via_pager; but blocked by
            # UnicodeDecodeError issue.
            print(tables)
            return
        if localized:
            return
        # If an output path was specified, there's no further need to init or
        # update catalogs.
        if o:
            return
        if init:
            text = 'Initializing {} empty translation catalogs.'
            pod.logger.info(text.format(len(locales)))
            catalogs.init(locales=locales, include_header=include_header)
            return
        if update:
            locales = validate_locales(catalogs.list_locales(), locale)
            text = 'Updating {} catalogs with extracted messages.'
            pod.logger.info(text.format(len(locales)))
            catalogs.update(locales=locales, include_header=include_header,
                            use_fuzzy_matching=use_fuzzy_matching,
                            include_obsolete=include_obsolete)
    return pod


def validate_locales(valid_locales, locales):
    """Validate that the locale is a valid local based on the list of locales."""
    for each in locales:
        if each not in valid_locales:
            text = ('{} is not a valid translation catalog locale. '
                    'Valid locales are: {}')
            valid_locales = [str(locale) for locale in valid_locales]
            raise ValueError(text.format(each, ', '.join(valid_locales)))
    return locales or valid_locales



class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class NotFoundError(Error, IOError):
    pass


class NoCatalogsError(Error):
    """Error when no catalogs are found for translations."""
    pass



"""Document fields for accessing the meta fields parsed from the document."""

import re
from grow.common import untag


LOCALIZED_KEY_REGEX = re.compile(r'(.*)@([^@]+)$')


class DocumentFields(object):

    def __contains__(self, item):
        return item in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __init__(self, data, locale_identifier=None, params=None):
        self._locale_identifier = locale_identifier
        self._params = params
        self._data = untag.Untag.untag(
            data, locale_identifier, params=params)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return "<DocumentFields(locale='{}')>".format(self._locale_identifier)

    def get(self, key, default=None):
        return self._data.get(key, default)

    def keys(self):
        return self._data.keys()

    def update(self, updated):
        updated = untag.Untag.untag(
            updated, self._locale_identifier, params=self._params)
        self._data.update(updated)



"""Google Translator toolkit for translating documents."""

import datetime
import base64
import httplib2
from urllib import parse as url_parse
from googleapiclient import discovery
from googleapiclient import errors
from grow.common import oauth
from grow.common import utils
from . import base


EDIT_URL_FORMAT = 'https://translate.google.com/toolkit/workbench?did={}'
GTT_DOCUMENTS_BASE_URL = 'https://www.googleapis.com/gte/v1/documents'
OAUTH_SCOPE = 'https://www.googleapis.com/auth/gte'
STORAGE_KEY = 'Grow SDK - Google Translator Toolkit'


class AccessLevel(object):
    ADMIN = 'ADMIN'
    READ_AND_COMMENT = 'READ_AND_COMMENT'
    READ_AND_WRITE = 'READ_AND_WRITE'
    READ_ONLY = 'READ_ONLY'


class ChangeType(object):
    MODIFY = 'MODIFY'
    ADD = 'ADD'


def raise_service_error(http_error, locale=None, ident=None):
    message = 'HttpError {} for {} returned "{}"'.format(
        http_error.resp.status, http_error.uri,
        http_error._get_reason().strip())
    raise base.TranslatorServiceError(message=message, locale=locale, ident=ident)


class Gtt(object):

    def __init__(self):
        self.service = discovery.build('gte', 'v1', http=self.http)

    @property
    def http(self):
        credentials = oauth.get_or_create_credentials(
            scope=OAUTH_SCOPE, storage_key=STORAGE_KEY)
        http = httplib2.Http()
        http = credentials.authorize(http)
        return http

    def get_document(self, document_id):
        # pylint: disable=no-member
        return self.service.documents().get(documentId=document_id).execute()

    def update_acl(self, document_id, acl_change):
        body = {}
        body['gttAclChange'] = acl_change
        # pylint: disable=no-member
        return self.service.documents().update(
            documentId=document_id, body=body).execute()

    def insert_document(self, name, content, source_lang, lang, mimetype,
                        acl=None, tm_ids=None, glossary_ids=None):
        content = base64.urlsafe_b64encode(content)
        doc = {
            'displayName': name,
            'gttAcl': acl,
            'language': lang,
            'mimetype': mimetype,
            'sourceDocBytes': content,
            'sourceLang': source_lang,
        }
        if acl:
            doc['gttAcl'] = acl
        if tm_ids:
            doc['tmIds'] = tm_ids
        if glossary_ids:
            doc['glossaryIds'] = glossary_ids
        try:
            # pylint: disable=no-member
            return self.service.documents().insert(body=doc).execute()
        except errors.HttpError as resp:
            raise_service_error(locale=lang, http_error=resp)

    def download_document(self, document_id):
        params = {
            'alt': 'media',
            'downloadContent': True,
        }
        url = '{}/{}?{}'.format(
            GTT_DOCUMENTS_BASE_URL,
            url_parse.quote(document_id), url_parse.urlencode(params))
        response, content = self.http.request(url)
        try:
            if response.status >= 400:
                raise errors.HttpError(response, content, uri=url)
        except errors.HttpError as resp:
            raise_service_error(ident=document_id, http_error=resp)
        return content


class GoogleTranslatorToolkitTranslator(base.Translator):
    KIND = 'google_translator_toolkit'
    has_immutable_translation_resources = True

    def _normalize_source_lang(self, source_lang):
        if source_lang is None:
            return 'en'
        source_lang = str(source_lang)
        source_lang = source_lang.lower()
        if source_lang == 'en_us':
            return 'en'
        return source_lang

    def _create_stat_from_gtt_response(self, resp, downloaded=False):
        url = EDIT_URL_FORMAT.format(resp['id'])
        lang = resp['language']
        source_lang = resp['sourceLang']
        stat = base.TranslatorStat(
            url=url,
            lang=lang,
            num_words=resp['numWords'],
            num_words_translated=resp['numWordsTranslated'],
            source_lang=source_lang,
            ident=resp['id'])
        if downloaded:
            stat.downloaded = datetime.datetime.now()
        else:
            stat.uploaded = datetime.datetime.now()
        return stat

    def _create_acl_from_config(self, config, lang, existing_acl=None):
        acl = None
        if 'acl' in config:
            acl = []
            for item in config['acl']:
                # Skip setting the ACL if it is localized and if its
                # locale doesn't match the catalog's locale.
                if 'locales' in item and lang not in item['locales']:
                    continue
                access_level = item.get('access_level',
                                        AccessLevel.READ_AND_WRITE)
                acl_item = {
                    'emailId': item['email'],
                    'accessLevel': access_level
                }
                if existing_acl:
                    change_type = ChangeType.ADD
                    for existing_item in existing_acl:
                        if existing_item['emailId'] == item['email']:
                            change_type = ChangeType.MODIFY
                    acl_item['type'] = change_type
                if 'can_reshare' in item:
                    acl_item['canReshare'] = item['can_reshare']
                acl.append(acl_item)
        return acl

    def _download_content(self, stat):
        gtt = Gtt()
        content = gtt.download_document(stat.ident)
        resp = gtt.get_document(stat.ident)
        stat = self._create_stat_from_gtt_response(resp, downloaded=True)
        return stat, content

    def _update_acl(self, stat, locale):
        gtt = Gtt()
        existing_doc = gtt.get_document(stat.ident)
        existing_acl = existing_doc['gttAcl']
        acl = self._create_acl_from_config(self.config, locale,
                                           existing_acl=existing_acl)
        if not acl:
            return
        gtt.update_acl(stat.ident, acl)

    def _update_meta(self, stat, locale, catalog):
        return self._update_acl(stat, locale)

    def _upload_catalog(self, catalog, source_lang, prune=False):
        gtt = Gtt()
        project_title = self.project_title
        name = '{} ({})'.format(project_title, str(catalog.locale))
        source_lang = self._normalize_source_lang(source_lang)
        lang = str(catalog.locale)
        glossary_ids = None
        tm_ids = self.config.get('tm_ids')
        glossary_ids = self.config.get('glossary_ids')
        acl = self._create_acl_from_config(self.config, lang)
        resp = gtt.insert_document(
            name=name,
            content=catalog.content,
            source_lang=str(source_lang),
            lang=lang,
            tm_ids=tm_ids,
            glossary_ids=glossary_ids,
            mimetype='text/x-gettext-translation',
            acl=acl)
        return self._create_stat_from_gtt_response(resp)



"""Profiling report for analyizing the performance of the app"""


class ProfileReport(object):
    """Analyzes the timers to report on the app timing."""

    def __init__(self, profile):
        self.profile = profile
        self.items = {}

        for timer in self.profile:
            if timer.key not in self.items:
                self.items[timer.key] = ReportItem(timer.key)
            self.items[timer.key].add_timer(timer)

    def analyze(self):
        """Performs an analysis of the timers."""
        pass

    def export(self):
        """Export the timer data for each timer."""
        exported = {}
        for key, item in self.items.items():
            exported[key] = item.export()
        return exported

    def pretty_print(self):
        """Prints out the report in a nice format."""
        for _, item in self.items.items():
            print('{} ({}): Avg {} Min {} Max {} '.format(
                item.key, len(item), item.average_duration, item.min_duration, item.max_duration))
            if len(item) > 1:
                for timer in item.top():
                    print(timer)


class ReportItem(object):
    """Report item used to store information about all timers with the same key."""

    def __init__(self, key):
        self.key = key
        self.timers = []
        self.start = None
        self.end = None

    def __len__(self):
        return len(self.timers)

    @property
    def average_duration(self):
        """Calculate the average duration of the timers."""
        duration_sum = 0
        for timer in self.timers:
            duration_sum = duration_sum + timer.duration
        return duration_sum / len(self.timers)

    @property
    def max_duration(self):
        """Calculate the maximum duration of the timers."""
        value = None
        for timer in self.timers:
            if not value or value < timer.duration:
                value = timer.duration
        return value

    @property
    def min_duration(self):
        """Calculate the minimum duration of the timers."""
        value = None
        for timer in self.timers:
            if not value or value > timer.duration:
                value = timer.duration
        return value

    @property
    def duration(self):
        """Duration of timers."""
        return self.end - self.start

    def add_timer(self, timer):
        """Add a timer to the report item to track it."""
        self.timers.append(timer)

        # Keep track of the first start and last end time.
        if not self.start or timer.start < self.start:
            self.start = timer.start
        if not self.end or timer.end > self.end:
            self.end = timer.end

    def export(self):
        """Export the timer data for each timer."""
        return {
            'start': self.start,
            'end': self.end,
            'timers': [t.export() for t in self.timers],
        }

    def top(self, count=5, ascending=False):
        """Get the top (or bottom) timers by duration."""
        top_list = sorted(self.timers, key=lambda k: k.duration)
        if not ascending:
            top_list.reverse()
        return top_list[:count]



"""Base class for the pre deploy hook."""

from grow.extensions.hooks import base_hook


class PreDeployHook(base_hook.BaseHook):
    """Hook for pre deploy."""

    KEY = 'pre_deploy'
    NAME = 'Pre Deploy'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, rendered_doc, command, *_args, **_kwargs):
        """Trigger the pre deploy hook."""
        if previous_result:
            return previous_result
        # Return None if nothing has changed.
        return None



"""Grow documents."""

import json
import logging
import os
import re
import slugify
import yaml
from grow.common import structures
from grow.common import untag
from grow.common import urls
from grow.common import utils
from grow.documents import document_fields
from grow.documents import document_format
from grow.pods import footnotes
from grow.translations import locales
from grow.pods import messages
from grow.translations import translation_stats


PATH_LOCALE_REGEX = re.compile(r'@([^-_]+)([-_]?)([^\.]*)(\.[^\.]+)$')
BUILT_IN_FIELDS = [
    'category',
    'date',
    'dates',
    'footnotes',
    'localization',
    'hidden',
    'order',
    'parent',
    'path',
    'slug',
    'sitemap',
    'title',
    'titles',
    'view',
]


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class DocumentExistsError(Error, ValueError):
    pass


class PathFormatError(Error, ValueError):
    pass


class Document(object):
    """Grow content document."""

    def __eq__(self, other):
        return (isinstance(self, Document)
                and isinstance(other, Document)
                and self.root_pod_path == other.root_pod_path)

    def __getattr__(self, name):
        if name == 'locale':
            return self._locale_kwarg
        try:
            return self.fields[name]
        except KeyError:
            return object.__getattribute__(self, name)

    def __ge__(self, other):
        self_id = (self.pod_path, str(self.locale))
        other_id = (other.pod_path, str(other.locale))
        return self_id >= other_id

    def __gt__(self, other):
        self_id = (self.pod_path, str(self.locale))
        other_id = (other.pod_path, str(other.locale))
        return self_id > other_id

    def __init__(self, pod_path, _pod, locale=None, _collection=None):
        self._locale_kwarg = locale
        utils.validate_name(pod_path)
        self.pod_path = pod_path
        # For multi-file localization and comparison.
        self.root_pod_path, _ = Document.parse_localized_path(pod_path)
        self.basename = Document._clean_basename(pod_path)
        self.base, self.ext = os.path.splitext(self.basename)
        self.pod = _pod
        self.collection = self._collection_exists(_collection)
        self._locale = utils.SENTINEL

    def __le__(self, other):
        self_id = (self.pod_path, str(self.locale))
        other_id = (other.pod_path, str(other.locale))
        return self_id <= other_id

    def __lt__(self, other):
        self_id = (self.pod_path, str(self.locale))
        other_id = (other.pod_path, str(other.locale))
        return self_id < other_id

    def __ne__(self, other):
        return self.pod_path != other.pod_path or self.pod != other.pod

    def __repr__(self):
        if self.locale:
            return "<Document({}, locale='{}')>".format(self.pod_path, self.locale)
        return "<Document({})>".format(self.pod_path)

    @classmethod
    def _clean_basename(cls, pod_path):
        base_pod_path = cls._locale_paths(pod_path)[-1]
        return os.path.basename(base_pod_path)

    @staticmethod
    def _collection_exists(collection):
        orig_collection = collection
        while not collection.exists:
            collection = collection.parent
            if not collection:
                collection = orig_collection
                break
        return collection

    @classmethod
    def _locale_paths(cls, pod_path):
        paths = [pod_path]
        parts = PATH_LOCALE_REGEX.split(pod_path)
        if len(parts) > 1:
            if parts[3]:  # [3] -> Country Code
                paths.append('{}@{}{}'.format(parts[0], parts[1], parts[4]))
            paths.append('{}{}'.format(parts[0], parts[4]))
        return paths

    @classmethod
    def clean_localized_path(cls, pod_path, locale):
        """Removed the localized part of the path."""
        if '@' in pod_path and locale is not None:
            base, _ = os.path.splitext(pod_path)
            if isinstance(locale, str):
                locale_str = locale
            else:
                locale_str = locale.alias or str(locale)
            if not base.endswith('@{}'.format(locale_str)):
                pod_path = cls._locale_paths(pod_path)[-1]
        return pod_path

    @classmethod
    def is_localized_path(cls, pod_path):
        return '@' in pod_path

    @classmethod
    def localize_path(cls, pod_path, locale):
        """Returns a localized path (formatted <base>@<locale>.<ext>) for
        multi-file localization."""
        pod_path = cls.clean_localized_path(pod_path, locale)
        if locale is None:
            return pod_path
        base, ext = os.path.splitext(pod_path)
        return '{}@{}{}'.format(base, locale, ext)

    @classmethod
    def parse_localized_path(cls, pod_path):
        """Returns a tuple containing the root pod path and the locale, parsed
        from a localized pod path (formatted <base>@<locale>.<ext>). If the
        supplied pod path does not contain a locale, the pod path is returned
        along with None."""
        groups = PATH_LOCALE_REGEX.split(pod_path)
        if len(groups) > 1:
            if groups[3]:
                locale = '{}{}{}'.format(groups[1], groups[2], groups[3])
            else:
                locale = groups[1]
            root_pod_path = '{}{}'.format(groups[0], groups[4])
            return root_pod_path, locale
        return pod_path, None

    def _init_locale(self, locale, pod_path):
        try:
            _, locale_from_path = Document.parse_localized_path(pod_path)
            if locale_from_path:
                locale = locale_from_path
            return self.pod.normalize_locale(
                locale, default=self.default_locale)
        except IOError as exc:  # Document does not exist.
            if '[Errno 2] No such file or directory' in str(exc):
                return None
            else:
                raise

    @property
    def body(self):
        return self.format.content if self.format.content else None

    @property
    def category(self):
        return self.fields.get('$category')

    @property
    def collection_base_path(self):
        """The base directory inside the collection."""
        return self.collection_sub_path[:-len(self.basename)]

    @property
    def collection_sub_path(self):
        """The pod path relative to the collection path."""
        return self.pod_path[len(self.collection.pod_path):]

    @property
    def collection_sub_path_clean(self):
        """The pod path relative to the collection path cleaned of locale."""
        return Document._locale_paths(
            self.pod_path[len(self.collection.pod_path):])[-1]

    @property
    def content(self):
        return self.format.raw_content

    @property
    def date(self):
        return self.fields.get('$date')

    @property
    def dates(self):
        """Built in field for dates."""
        return structures.AttributeDict(self.fields.get('$dates', {}))

    @utils.cached_property
    def default_locale(self):
        """Default document locale."""
        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            localization = self.collection.routes.localization(
                self.collection_sub_path, {})
        else:
            localization = self.format.front_matter.data.get(
                '$localization', {})

        if localization and 'default_locale' in localization:
            return self.pod.normalize_locale(localization['default_locale'])
        return self.collection.default_locale

    @property
    def exists(self):
        return self.pod.file_exists(self.pod_path)

    @utils.cached_property
    def fields(self):
        locale_identifier = str(self._locale_kwarg or self.default_locale)
        return document_fields.DocumentFields(
            self.format.front_matter.data, locale_identifier,
            params={
                'env': untag.UntagParamRegex(self.pod.env.name),
                'locale': untag.UntagParamLocaleRegex.from_pod(
                    self.pod, self.collection),
            })

    @utils.cached_property
    def footnotes(self):
        # Configure the footnotes based on the doc or podspec settings.
        footnote_config = self.fields.get(
            '$footnotes', self.pod.podspec.fields.get('footnotes', {}))
        locale = str(self.locale) if self.locale else None
        symbols = footnote_config.get('symbols', None)
        use_numeric_symbols = footnote_config.get('use_numeric_symbols', None)
        use_numeric = footnote_config.get('use_numeric', None)
        numeric_locales_pattern = footnote_config.get(
            'numeric_locales_pattern', None)
        return footnotes.Footnotes(
            locale, symbols=symbols, use_numeric=use_numeric,
            use_numeric_symbols=use_numeric_symbols,
            numeric_locales_pattern=numeric_locales_pattern)

    @utils.cached_property
    def format(self):
        return document_format.DocumentFormat.from_doc(doc=self)

    @property
    def formatted(self):
        return self.format.formatted

    @property
    def hidden(self):
        return self.fields.get('$hidden', False)

    @property
    def html(self):
        return self.formatted

    @utils.cached_property
    def locale(self):
        if self._locale is utils.SENTINEL:
            self._locale = self._init_locale(self._locale_kwarg, self.pod_path)
        return self._locale

    @property
    def hreflang(self):
        # https://en.wikipedia.org/wiki/Hreflang
        # hreflang values are used in sitemap.xml and in <meta> tags on a page.
        locale = self.locale
        if locale == self.default_locale:
            return 'x-default'
        value = str(locale)
        value = value.lower()
        value = value.replace('_', '-')
        return value

    @utils.cached_property
    def locale_safe(self):
        # During the initialization of the document the locale is used,
        # but the fields cannot be used to modify the locale. This is a 'safe'
        # way of getting the locale defined in the constructor with a fallback
        # to the collection default locale.
        return self._locale_kwarg or self.collection.default_locale

    @utils.cached_property
    def locale_paths(self):
        return Document._locale_paths(self.pod_path)

    @utils.cached_property
    def locales(self):
        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            localization = self.collection.routes.localization(
                self.collection_sub_path, {})
        else:
            # Default to none to be able to override the upstream locales.
            localization = self.fields.get('$localization', {})

        # Disable localization with $localization:~.
        if localization is None:
            return []
        if 'locales' in localization:
            codes = localization['locales'] or []
            return self.pod.normalize_locales(
                locales.Locale.parse_codes(codes))
        return self.collection.locales

    @property
    def order(self):
        return self.fields.get('$order', 0)

    @property
    @utils.memoize
    def parent(self):
        parent_pod_path = self.fields.get('$parent')
        if not parent_pod_path:
            return None
        return self.collection.get_doc(parent_pod_path, locale=self.locale)

    @property
    @utils.memoize
    def path_format(self):
        """Path format for current document."""
        if (self.locale
                and self.locale != self.default_locale
                and self.path_format_localized is not None):
            return self.path_format_localized
        return self.path_format_base

    @property
    @utils.memoize
    def path_format_base(self):
        """Path format for base document."""
        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            return self.collection.routes.path(
                self.collection_sub_path, self.collection.path_format)
        return self.fields.get('$path', self.collection.path_format)

    @property
    @utils.memoize
    def path_format_localized(self):
        """Path format for localized documents."""
        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            localization = self.collection.routes.localization(
                self.collection_sub_path, {})
        else:
            localization = self.fields.get('$localization', {})

        if 'path' in localization:
            return localization['path']

        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            path_format = self.collection.routes.path(self.collection_sub_path, '')
        else:
            path_format = self.fields.get('$path', '')

        if path_format and '{locale}' in path_format:
            return path_format
        col_localization = self.collection.localization
        if col_localization and 'path' in col_localization:
            return col_localization.get('path')
        col_path_format = self.collection.path_format
        if col_path_format and '{locale}' in col_path_format:
            return col_path_format
        return None

    @property
    @utils.memoize
    def path_params(self):
        """Path params for current document."""
        params = {}

        # TODO: Allow for defining custom path param options in the
        # doc or collection.

        return params

    @property
    @utils.memoize
    def path_params_localized(self):
        """Path params for current document."""
        params = {}

        # TODO: Allow for defining custom path param options in the
        # doc or collection.

        # When there are locales in a path enumerate the possible locales.
        if '{locale}' in self.path_format_localized:
            locale_values = set()
            for locale in self.locales:
                locale.set_alias(self.pod)
                locale_values.add(locale.alias)
                locale_values.add(locale.alias.lower())
            params['locale'] = locale_values

        return params

    @property  # Cached in document format.
    def raw_content(self):
        return self.format.raw_content

    @property
    def slug(self):
        value = self.fields.get('$slug')
        if value:
            return value
        if not self.title:
            return None
        if self.pod.is_enabled(self.pod.FEATURE_OLD_SLUGIFY):
            return utils.slugify(self.title)
        return slugify.slugify(self.title)

    @property
    def sitemap(self):
        return self.fields.get('$sitemap')

    @property
    def title(self):
        return self.fields.get('$title')

    @utils.cached_property
    def translation_stats(self):
        return translation_stats.TranslationStats()

    @property
    def url(self):
        path = self.get_serving_path()
        if path:
            return urls.Url(
                path=path,
                host=self.pod.env.host,
                port=self.pod.env.port,
                scheme=self.pod.env.scheme)

    @property
    def view(self):
        if self.collection.features(self.collection.FEATURE_SEPARATE_ROUTING):
            view_format = self.collection.routes.view(
                self.collection_sub_path, self.collection.view)
        else:
            view_format = self.fields.get('$view', self.collection.view)
        if view_format is not None:
            return self.pod.path_format.format_view(
                self, view_format)
        return ''

    def delete(self):
        self.pod.delete_file(self.pod_path)

    def get_date(self, date_name=None):
        if date_name is None:
            return self.date
        return self.dates.get(date_name, self.date)

    @utils.memoize
    def has_serving_path(self):
        return bool(self.path_format)

    def inject(self, fields=utils.SENTINEL, body=utils.SENTINEL):
        """Injects without updating the copy on the filesystem."""
        if fields != utils.SENTINEL:
            self.fields.update(fields)
        if body != utils.SENTINEL:
            self.format.body = body
        self.pod.logger.info('Injected -> {}'.format(self.pod_path))

    @utils.memoize
    def get_serving_path(self):
        # Get root path.
        path_format = self.path_format
        if path_format is None:
            raise PathFormatError(
                'No path format found for {}. You must specify a path '
                'format in either the blueprint or the document.'.format(self))
        path_format = (path_format
                       .replace('<grow:locale>', '{locale}')
                       .replace('<grow:slug>', '{slug}'))

        # Handle default date formatting in the url.
        while '{date|' in path_format:
            re_date = r'({date\|(?P<date_format>[a-zA-Z0-9_%-]+)})'
            match = re.search(re_date, path_format)
            if match:
                formatted_date = self.date
                formatted_date = formatted_date.strftime(
                    match.group('date_format'))
                path_format = (path_format[:match.start()] + formatted_date +
                               path_format[match.end():])
            else:
                # Does not match expected format, let the normal format attempt
                # it.
                break

        # Handle the special formatting of dates in the url.
        while '{dates.' in path_format:
            re_dates = r'({dates\.(?P<date_name>\w+)(\|(?P<date_format>[a-zA-Z0-9_%-]+))?})'
            match = re.search(re_dates, path_format)
            if match:
                formatted_date = self.get_date(match.group('date_name'))
                date_format = match.group('date_format') or '%Y-%m-%d'
                formatted_date = formatted_date.strftime(date_format)
                path_format = (path_format[:match.start()] + formatted_date +
                               path_format[match.end():])
            else:
                # Does not match expected format, let the normal format attempt
                # it.
                break

        try:
            return self.pod.path_format.format_doc(
                self, path_format, locale=self.locale)
        except KeyError:
            logging.error('Error with path format: {}'.format(path_format))
            raise

    @utils.memoize
    def get_serving_path_base(self):
        """Get the base (default locale) serving path."""
        return self.pod.path_format.format_doc(
            self, self.path_format_base, locale=self.default_locale)

    @utils.memoize
    def get_serving_path_localized(self):
        """Get the serving paths for each non-default locale."""
        return self.pod.path_format.format_doc(
            self, self.path_format_localized, parameterize=True)

    @utils.memoize
    def get_serving_paths_localized(self):
        """Get the serving paths for each non-default locale."""
        paths = {}
        for locale in self.locales:
            if locale == self.default_locale:
                continue
            paths[locale] = self.pod.path_format.format_doc(
                self, self.path_format_localized, locale=locale)
        return paths

    def localize(self, locale):
        return self.collection.get_doc(self.root_pod_path, locale=locale)

    def next(self, docs=None):
        """Finds the next document relative to the current doc."""
        if docs is None:
            docs = self.collection.list_docs()
        for i, doc in enumerate(docs):
            if type(doc) != self.__class__:
                raise ValueError('Usage: {{doc.next(<docs>)}}.')
            if doc == self:
                n = i + 1
                if n == len(docs):
                    return None
                return docs[i + 1]

    def prev(self, docs=None):
        """Finds the preceding document relative to the current doc."""
        if docs is None:
            docs = self.collection.list_docs()
        prev_doc = None
        for doc in docs:
            if type(doc) != self.__class__:
                raise ValueError('Usage: {{doc.prev(<docs>)}}.')
            if doc == self:
                return prev_doc
            prev_doc = doc
        return None

    def titles(self, title_name=None):
        if title_name is None:
            return self.title
        titles = self.fields.get('$titles', {})
        return titles.get(title_name, self.title)

    def to_message(self):
        message = messages.DocumentMessage()
        message.basename = self.basename
        message.pod_path = self.pod_path
        message.collection_path = self.collection.collection_path
        message.body = self.body
        message.content = self.content
        message.fields = json.dumps(self.fields, cls=utils.JsonEncoder)
        message.serving_path = self.get_serving_path()
        return message

    def write(self, fields=utils.SENTINEL, body=utils.SENTINEL):
        if body is utils.SENTINEL:
            body = self.body if self.exists else ''
        self.format.update(fields=fields, content=body)
        new_content = self.format.to_raw_content()
        self.pod.write_file(self.pod_path, new_content)


# Allow the yaml dump to write out a representation of the document.
def doc_representer(dumper, data):
    return dumper.represent_scalar('!g.doc', data.pod_path)


yaml.SafeDumper.add_representer(Document, doc_representer)



"""Colors for use in logging."""

import colored

# Colors and styles that look good on both black and white backgrounds.
CAUTION = colored.fg('yellow') + colored.attr("bold")
EMPHASIS = colored.fg('dark_orange_3a')
ERROR = colored.fg('red') + colored.attr("bold")
HIGHLIGHT = colored.fg('light_green')
SUCCESS = colored.fg('green') + colored.attr("bold")

# pylint: disable=invalid-name
stylize = colored.stylize



"""Base class for the podspec static_dir hook."""

from grow.extensions.hooks import base_hook


class PodspecStaticDirHook(base_hook.BaseHook):
    """Hook for podspec static_dir."""

    KEY = 'static_dir'
    NAME = 'Podspec Static Dir'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, *_args, **_kwargs):
        """Trigger the podspec static_dir hook."""
        if previous_result:
            return previous_result
        return []



"""Partials contain organized units for templating."""

import os
from grow.documents import document_front_matter
from grow.common import utils


split_front_matter = document_front_matter.DocumentFrontMatter.split_front_matter


class Partial(object):
    """Partial from the partial directory."""

    def __init__(self, pod_path, pod):
        self.pod_path = self.clean_path(pod_path)
        self.key = os.path.basename(self.pod_path)
        self.pod = pod

    def __repr__(self):
        return '<Partial "{}">'.format(self.key)

    @classmethod
    def clean_path(cls, pod_path):
        """Clean up the pod path for the partials."""
        return pod_path.rstrip('/')

    @property
    def config(self):
        if not self.exists:
            return {}

        front_matter, _ = split_front_matter(
            self.pod.read_file(self.template_pod_path))
        if not front_matter:
            return {}
        return utils.parse_yaml(front_matter, pod=self.pod)

    @property
    def exists(self):
        return self.pod.file_exists(self.pod_path)

    @property
    def template_pod_path(self):
        return os.path.join(self.pod_path, '{}.html'.format(self.key))


class ViewPartial(object):
    """Partial from the views directory."""

    def __init__(self, pod_path, pod):
        self.pod_path = self.clean_path(pod_path)
        base, _ = os.path.splitext(self.pod_path)
        self.key = os.path.basename(base)
        self.pod = pod

    def __repr__(self):
        return '<ViewPartial "{}">'.format(self.key)

    @classmethod
    def clean_path(cls, pod_path):
        """Clean up the pod path for the partial."""
        return pod_path

    @property
    def config(self):
        if not self.exists:
            return {}

        front_matter, _ = split_front_matter(self.pod.read_file(
            self.template_pod_path))
        if not front_matter:
            return {}
        return utils.parse_yaml(front_matter, pod=self.pod)

    @property
    def exists(self):
        return self.pod.file_exists(self.pod_path)

    @property
    def template_pod_path(self):
        return self.pod_path



import git
from grow.common import utils
from grow.deployments import messages


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class NoGitHeadError(Error, ValueError):
    pass


def create_commit_message(repo):
    message = messages.CommitMessage()
    try:
        commit = repo.head.commit
    except ValueError:
        raise NoGitHeadError('On initial commit, no HEAD yet.')
    try:
        repo.git.diff('--quiet')
        has_unstaged_changes = False
    except git.GitCommandError:  # pragma: no cover
        has_unstaged_changes = True
    message.has_unstaged_changes = has_unstaged_changes
    message.sha = commit.hexsha
    message.message = commit.message
    try:
        message.branch = repo.head.ref.name
    except TypeError:
        # Allow operating in an environment with a detached HEAD.
        pass
    message.author = messages.AuthorMessage(
        name=commit.author.name, email=commit.author.email)
    return message



from protorpc import messages
from protorpc import message_types


class AuthorMessage(messages.Message):
    name = messages.StringField(1)
    email = messages.StringField(2)


class CommitMessage(messages.Message):
    sha = messages.StringField(1)
    author = messages.MessageField(AuthorMessage, 2)
    date = message_types.DateTimeField(3)
    message = messages.StringField(4)
    has_unstaged_changes = messages.BooleanField(5)
    branch = messages.StringField(6)


class FileMessage(messages.Message):
    path = messages.StringField(1)
    sha = messages.StringField(2)
    deployed = message_types.DateTimeField(3)
    deployed_by = messages.MessageField(AuthorMessage, 4)
    action = messages.StringField(5)


class IndexMessage(messages.Message):
    files = messages.MessageField(FileMessage, 1, repeated=True)
    deployed = message_types.DateTimeField(4)
    deployed_by = messages.MessageField(AuthorMessage, 5)
    commit = messages.MessageField(CommitMessage, 6)


class WhatChangedMessage(messages.Message):
    commit = messages.MessageField(CommitMessage, 1)
    files = messages.MessageField(FileMessage, 2, repeated=True)


class DiffMessage(messages.Message):
    adds = messages.MessageField(FileMessage, 1, repeated=True)
    edits = messages.MessageField(FileMessage, 2, repeated=True)
    deletes = messages.MessageField(FileMessage, 3, repeated=True)
    nochanges = messages.MessageField(FileMessage, 4, repeated=True)
    indexes = messages.MessageField(IndexMessage, 5, repeated=True)
    what_changed = messages.StringField(6)
    is_partial = messages.BooleanField(7)


class FileCountMessage(messages.Message):
    ext = messages.StringField(1)
    count = messages.IntegerField(2)


class StatsMessage(messages.Message):
    num_collections = messages.IntegerField(101)
    num_documents = messages.IntegerField(102)
    num_static_files = messages.IntegerField(103)
    num_files_per_type = messages.MessageField(FileCountMessage, 104, repeated=True)

    locales = messages.StringField(123, repeated=True)
    langs = messages.StringField(124, repeated=True)
    num_messages = messages.IntegerField(125)
    num_untranslated_messages = messages.IntegerField(126)
